/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <osbind.h>
#include <fcntl.h>
#include <ioctl.h>
#include <mintbind.h>
#include <signal.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "app_man.h"
#include "xa_codes.h"
#include "c_window.h"
#include RSCHNAME
#include "desktop.h"
#include "objects.h"
#include "widgets.h"
#include "menuwidg.h"
#include "xa_rsrc.h"

/*
 *	Application Manager
 *	This module handles various ways of swapping the current application...
*/

short dead_exit_contrl[] = {XA_CLIENT_EXIT, 0, 0, 0, 0};
AESPB dead_client_pb;

/*
 *	Go round and check whether any clients have died & we didn't notice.
 *	- Useful for cleaning up clients that weren't launched via a shell_write.
 */
void find_dead_clients(LOCK lock)
{
/* FIX_PC use of r eliminated; int?long? who wants to know? */
/* HR: use proper loop. */
	XA_CLIENT *client = client_list;
	while (client)
	{
		short pid = client->pid;
		if (pid > 0 && pid != AESpid)
	/* HR: We dont look at client_end, because it is possible that a app is started by shell_write,
	    but never performed appl_init. */
			if (Pkill(pid, SIGNULL))	/* Poll to see if the client has died since we last looked... */
			{
				XA_CLIENT *next;
				Pwaitpid(-1, 1, nil);
	
				DIAGS(("Dead client %s(%d) - cleaning up\n",client->proc_name, pid));
			
				dead_client_pb.contrl = dead_exit_contrl;	/* If client is dead, send ourselves a message to clean up.... */
				if (client->init)
					XA_client_exit(lock, client, &dead_client_pb);	/* Run the application exit cleanup */

				if (client->kernel_end)
				{
					Fclose(client->kernel_end);
					client_handle_mask &= ~(1L << client->kernel_end);
					DIAGS(("DEAD: Closed kernel_end %d\n", client->kernel_end));
				}
				DIAGS(("DEAD: Free client: %s(%d)\n",client->proc_name,pid));

				next = client->next;			/* HR before FreeClient :-) */
				FreeClient(lock, client);
				client = next;
				continue;
			}
		client = client->next;
	}
}

/*
 *	Attempt to recover a system that has locked up
 */
void recover(void)
{	
	DIAG((D.appl,-1,"Attempting to recover control.....\n"));

	if ((update_lock) && (update_lock != AESpid))
	{
		DIAG((D.appl,-1,"Killing owner of update lock\n"));
		Pkill(update_lock, SIGKILL);
	}
	
	if (((mouse_lock) && (mouse_lock != update_lock)) && (mouse_lock != AESpid))
	{
		DIAG((D.appl,-1,"Killing owner of mouse lock\n"));
		Pkill(mouse_lock, SIGKILL);
	}
	
	v_show_c(V_handle, 0);
}

/*
 *	Swap the main root window's menu-bar to be the next application's
 *	(cycles which app owns the main menu-bar).
 *	NOTE: This only swaps the menu-bar, it doesn't swap the topped window.
 *
 *  HR: static pid array.
 *  See also click_menu_widget() for APP's
 */

void swap_menu(LOCK lock, short pid)
{
	XA_CLIENT *client = Pid2Client(pid);
	XA_WIDGET_TREE *menu_bar = root_window->widgets[XAW_MENU].stuff;
DIAG((D.menu,-1,"swap_menu to '%s'(%d)\n", client->name, pid));

	if (client->std_menu.tree)		/* HR: If the client has no menu bar, no need for a change */
	{

/* HR menu widget.tree */
		if (client->std_menu.tree != menu_bar->tree)		/* Different menu? */
		{
		/* HR menu widget the following not needed anymore */
			if (client->std_menu.owner == AESpid)
			{
DIAG((D.menu,-1,"swapped to AES\n"));
				*menu_bar = Pid2Client(AESpid)->std_menu;	/* Apps with no menu get the default menu, and that's owned by the AES */
			}
			else
			{
DIAG((D.menu,-1,"swapped to pid %d\n",pid));
				*menu_bar = client->std_menu;	/* Apps with menus own their own menu */
			}
		}
		else
		{
DIAG((D.menu,-1,"Same menu %d\n",pid));
		}
	}

	if (   client->desktop						/* Change desktops? */
		&& client->desktop != desktop
		&& client->desktop != ResourceTree(system_resources, DEF_DESKTOP)
		)
	{
		set_desktop(client->desktop);
		root_window->owner = pid;
		display_non_topped_window(lock, 30, root_window, NULL);
	}
	else
	if (client->std_menu.tree)	/* No - just change menu bars */
	{
		RECT clip;
		rp_2_ap(root_window, root_window->widgets + XAW_MENU, &clip);
		v_hide_c(V_handle);
		set_clip(&clip);
		display_menu_widget(lock, root_window, &root_window->widgets[XAW_MENU]);
		v_show_c(V_handle, 1);
	}
}

void next_app_menu(LOCK lock)
{
	XA_WIDGET_TREE *menu_bar = root_window->widgets[XAW_MENU].stuff;
	XA_CLIENT *client;
	short next_app, start_app;
	
	start_app = menu_bar->owner;
	
	next_app = start_app + 1;
	
	do
	{
		if (next_app >= MAX_PID)
			next_app = 0;
		client = Clients[next_app];
		if (client)
		{
			if ((client->client_end) || (next_app == AESpid))		/* Valid client ? */
			{
				swap_menu(lock, next_app);
				return;
			}
		}
		next_app++;
	} while (next_app != start_app);
}

void list_apps(void)
{
	short f;
	
	DIAGS(("========== XaAES Client Applications List ===========\n"));
	/* HR: static pid array */
	for (f = 0; f < MAX_PID; f++)
	{
		XA_CLIENT *client = Clients[f];
		if (client && client->client_end)
		{
			DIAGS(("%d:%s..Wait[", f, client->name));
			if (client->waiting_for & XAWAIT_BUTTON)
				DIAGS(("B"));
			if (client->waiting_for & XAWAIT_KEY)
				DIAGS(("K"));
			if (client->waiting_for & XAWAIT_MOUSE1)
				DIAGS(("1"));
			if (client->waiting_for & XAWAIT_MOUSE2)
				DIAGS(("2"));
			if (client->waiting_for & XAWAIT_MESSAGE)
				DIAGS(("M"));
			if (client->waiting_for & XAWAIT_TIMER)
				DIAGS(("T"));
			if (client->waiting_for & XAWAIT_CHILD)
				DIAGS(("C"));
			if (client->waiting_for & XAWAIT_MULTI)
				DIAGS(("{MULTI}"));
			DIAGS(("]:AESparent=%d", client->parent));
			if (client->std_menu.tree)		/* HR: now widget */
				DIAGS((":HasMenubar"));
			if (((XA_WIDGET_TREE *)root_window->widgets[XAW_MENU].stuff)->owner == f)
				DIAGS(("(current)"));
			if (client->desktop)
				DIAGS((":OwnsDesktop"));
			if (client->msg)
				DIAGS((":MsgPending"));
			if (mouse_lock == f)
				DIAGS((":OwnsMouseLock"));
			if (update_lock == f)
				DIAGS((":OwnsUpdateLock"));
			
			DIAGS(("\n"));
		}
	}
}

void status_dump(void)
{
	DIAGS(("========== XaAES Status ===========\n"));

	DIAGS(("AESpid=%d\n", AESpid));
	DIAGS(("DSKpid=%d\n", DSKpid));

	DIAGS(("Display Device: P_handle=%d, V_handle=%d\n", P_handle, V_handle));
	DIAGS((" size=[%d,%d], colours=%d, bitplanes=%d\n", display.r.w, display.r.h, display.colours, display.planes));

	DIAGS(("mouse_lock=%d, mouse_cnt=%d\n", mouse_lock, mouse_cnt));
	DIAGS(("update_lock=%d, update_cnt=%d\n", update_lock, update_cnt));
	
	DIAGS(("root window menu : owner=%d\n", ((XA_WIDGET_TREE *)root_window->widgets[XAW_MENU].stuff)->owner));
}
