/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <fcntl.h>
#include <ioctl.h>
#include <osbind.h>
#include <mintbind.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "c_window.h"
#include RSCHNAME
#include "rectlist.h"
#include "desktop.h"
#include "menuwidg.h"
#include "xa_rsrc.h"


/*
 *	Open the clients comms pipe in response to an XA_NEW_CLIENT message
 */
AES_function XA_new_client	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	char pipe_name[50];
	char fmt[] = "u:" sbslash "pipe" sbslash "XaClnt.%d\0";
	short f;

DIAG((D.appl,-1,"XA_new_client for %d\n",client->pid));

DIAG((D.appl,-1," - client_end %d\n",client->client_end));
	if (!client->client_end)	/* If this occurs, then we've got a problem */
	{
		DIAG((D.appl,-1,"Init client - Error: client pipe does not exist yet?\n"));
		/* PANIC - opening a global handle won't help because global
			handles can't be used in an Fselect mask... */
		return XAC_BLOCK ;
	}

DIAG((D.appl,-1," - kernel_end %d\n",client->kernel_end));
	if (!client->kernel_end)
	{
		sprintf(pipe_name, fmt, client->pid);						/* Open the clients reply pipe for writing to */
		client->kernel_end = Fopen(pipe_name, O_RDWR|O_DENYNONE);		/* Kernal's end of pipe*/
		client_handle_mask |= 1L << (client->kernel_end);
	}

#if 0		/* HR: std_menu is now a complete widget_tree :-) */
	client->std_menu = ResourceTree(system_resources, SYSTEM_MENU);
#else
	client->std_menu = Pid2Client(AESpid)->std_menu;
#endif

	if (*client->cmd_name)
	{
		char *q = client->cmd_name,
			 *p = strrchr(q,bslash);
		if (!p)						/* no '\\' */
			p = strrchr(q,slash);
		if (!p)						/* no '/' */
			p = strrchr(q,':');
		if (!p)						/* no ':' */
			p = q-1;
		q = strchr(q,'.');
		if (q)
			*q = 0;
		strncpy(client->proc_name,p+1,8);
		client->proc_name[8] = '\0';
		strnupr(client->proc_name,8);
	}
	sprintf(client->name, "  %s", client->proc_name);  /* awaiting menu_register */

	DIAG((D.appl,-1,"Init client '%s' pid=%d, client_end=%d, kernel_end=%d\n", client->proc_name, client->pid, client->client_end, client->kernel_end));

	client->init = true;
	return XAC_DONE;	/* We now unblock the client, 'coz we've setup our end */
}

/*
 *	Close down the client reply pipe in response to an XA_CLIENT_EXIT message
 *     HR: No!! it doesnt!
 *	- Also does a tidy-up and deletes all the clients windows (in case some untidy programs
 *	  fail to close them for themselves).
 *	- Also disposes of any pending messages.
*/
AES_function XA_client_exit	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_AESMSG_LIST *nm, *dnm;
	XA_WIDGET_TREE *menu_bar = root_window->widgets[XAW_MENU].stuff;
	short owner_top_wm;

	DIAG((D.appl,-1,"XA_client_exit: %s(%d)\n",client->proc_name, client->pid));

/* Go through and check that all windows belonging to this client are */
	owner_top_wm = wind_new(lock, client->pid);

/* Dispose of any pending messages for the client */
	while (client->msg)		/* HR */
	{
		nm = client->msg->next;
		free(client->msg);
		client->msg = nm;
	}

/* If the client forgot to remove its menu bar, better do it now */
	if (menu_bar->tree == client->std_menu.tree)
	{
/* HR: std_menu is now widget tree */
		if (owner_top_wm != -1)
			*menu_bar = Pid2Client(owner_top_wm)->std_menu;
		else
			*menu_bar = Pid2Client(AESpid)->std_menu;
		v_hide_c(V_handle);
		display_menu_widget(lock, root_window, &root_window->widgets[XAW_MENU]);
		v_show_c(V_handle, 1);
	}

/* Did the exiting app forget to remove a custom desktop? */
	if ((desktop == client->desktop) && (desktop != ResourceTree(system_resources, DEF_DESKTOP)))
	{
		set_desktop(Pid2Client(menu_bar->owner)->desktop);
		root_window->owner = menu_bar->owner;

		display_non_topped_window(lock, 70, root_window, NULL);
	}


/* If the client forgot to free its resources, we do it for them. */
	XA_free_all(client);		/* HR: free all blocks allocated on behalf of the client. */
	client->std_resource = NULL;

	/* Free command tail and name *only if* they were malloced: */
	/* HR dont compare with a pointer!!!! */
	if (client->heap_tail)		/* HR what if another constant was used? */
		free(client->cmd_tail);
	if (client->heap_name)
		free(client->cmd_name);
	
	if (update_lock == client->pid)
	{					/* Unlock mouse & screen */
		update_lock = 0;
		update_cnt = 0;
	}

	if (mouse_lock == client->pid)
	{
		mouse_lock = FALSE;
		mouse_cnt = 0;
	}

	/* HR: FreeClient() theoratically destroys all information about a client. */
	/* so we postpone the physical destruction until SIGCHILD */

#if 0	/* HR: See also XA_appl_exit; Here it is under the wrong pid. :-) */
	if (client->client_end)
	{
		Fclose(client->client_end);
		client->client_end = 0;  	/* Do not reenter */
	}
#endif

	top_next_window(lock);		/* HR */

	client->init = false;

	return XAC_DONE;	 	/* Closed down, let the client move on & exit */
}
