/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <memory.h>
#include <string.h>
#include "mintbind.h"

#include "xa_types.h"
#include "xa_globl.h"

#include "xa_evnt.h"
#include "xa_mouse.h"
#include "rectlist.h"
#include "objects.h"
#include "kernal.h"

static
short pending_msgs(LOCK lock, XA_CLIENT *client, AESPB *pb, short type)
{
	short rtn;

	if unlocked(clients)				/* I dont think these are needed. */
		Sema_Up(CLIENTS_SEMA);

	/* Is there a widget still active (like a scroll arrow)? If so, check with the action first
	 * as it may result in some messages (just in case we've not got any already)
	 */
	 
	if (!client->msg)
	{
		if (client->widget_active)
		{
			WidgetBehaviour *wc, *disp;
			XA_PENDING_WIDGET *pending;
			XA_WIDGET *widg;
			XA_WINDOW *wind;
			pending = client->widget_active;
			wc = pending->action;
			widg = pending->widg;
			wind = pending->wind;

			rtn = (*wc)(lock|clients, wind, widg);	/* Call the pending action */

			if (rtn)	/* If the widget click/drag function returned TRUE we reset the state of the widget */
			{
				XA_RECT_LIST *rl = rect_get_system_first(wind);

				widg->stat = NONE;					/* Flag the widget as de-selected */
				disp = widg->behaviour[XACB_DISPLAY];		/* get the redraw function for this widget */

				v_hide_c(V_handle);
				while (rl)	/* Walk the rectangle list */
				{
					set_clip(&rl->r);
					(*disp)(lock|clients, wind, widg);
					rl = rect_get_system_next(wind);
				}
				v_show_c(V_handle, 1);
			}
		}
	}

	if (client->msg)	/* Are there any messages pending? */
	{
		MSG_BUF *buf = (MSG_BUF *)(pb->addrin[0]);
		XA_AESMSG_LIST *msg = client->msg;
		client->msg = msg->next;
	
		*buf = msg->message;				/* Copy the message into the clients buffer */
		multi_intout(pb->intout, type);		/* HR: fill out mouse data!!! */
		free(msg);
		if unlocked(clients)
			Sema_Dn(CLIENTS_SEMA);
#if GENERATE_DIAGS
		if (buf->m[0] == WM_REDRAW)
		{
			DIAG((D.wind, -1, "pending WM_REDRAW got for %s(%d) %d/%d,%d/%d\n",
				client->name, client->pid, buf->m[4], buf->m[5], buf->m[6], buf->m[7]));
		}
#endif
		return TRUE;		/* Return TRUE to unblock the client */
	}
	
	client->waiting_for = XAWAIT_MESSAGE;	/* Mark the client as waiting for messages */
	client->waiting_pb = pb;
	if unlocked(clients)
		Sema_Dn(CLIENTS_SEMA);
	return FALSE;
}

bool pending_key_strokes(AESPB *pb, short type)
{
	if (pending_keys.cur != pending_keys.end)			/* HR: keys may be queued */
	{
		short key = pending_keys.k[pending_keys.cur++];
		if (pending_keys.cur == sizeof(pending_keys.k)/sizeof(short))
			pending_keys.cur = 0;
		if (type == MU_KEYBD)
		{
			multi_intout(pb->intout, type);
			pb->intout[5] = key;
		}
		else
			pb->intout[0] = key;
		DIAG((D.keybd, -1, "pending key: %x\n", key));
		return true;
	}
	return false;
}

/*
 *	The essential evnt_multi() call
 */

#if GENERATE_DIAGS
char *xev[] = {"KBD","BUT","M1","M2","MSG","TIM","6","7","8","9","10","11","12","13","14","15"};
void show_ev(unsigned short ev, char *x)
{
	short i=0;
	*x=0;
	while(ev)
	{
		if (ev&1)
			strcat(x,xev[i]), strcat(x,"|");
		ev>>=1;
		i++;
	}
	i = strlen(x)-1;
	if (x[i] == '|')
		x[i] = 0;
}
#endif

AES_function XA_evnt_multi	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short events = pb->intin[0];
	unsigned long ret = XAC_DONE;
	short f, rtn, new_waiting_for = 0;
#if GENERATE_DIAGS
	char evtxt[128];
#endif

	CONTROL(16,7,1)

	client->waiting_pb = pb;	/* Store a pointer to the AESPB to fill when the event(s) */
								/*  finally arrive. */

/* HR */
#if GENERATE_DIAGS
	show_ev(events,evtxt);
	DIAG((D.multi,client->pid,"evnt_multi for %d, evnt=%s, clks=0x%x, msk=0x%x, bst=x%x\n",
			client->pid,evtxt,pb->intin[1],pb->intin[2],pb->intin[3]));
#endif
/*
	Excerpt from nkcc.doc, courtesy Harald Siegmund:

	Note: the NKCC button event handler supports the (undocumented)
	negation flag, which is passed in bit 8 of the parameter <bclicks>
	(maximum # of mouse clicks to wait for). You don't know this flag?
	I found an article about it in the c't magazine (I think it was
	issue 3/90, or maybe 4/90??) - and I damned Atari for their bad
	documentation. This flag opens the way to check BOTH mouse buttons
	at the same time without any problems. When set, the return
	condition is inverted. Let's have a look at an example:
	
	mask = evnt_multi(MU_BUTTON,2,3,3,...
	
	This doesn't work the way we want: the return condition is
	"button #0 pressed AND button #1 pressed". But look at this:
	
	mask = evnt_multi(MU_BUTTON,0x102,3,0,...
	
	Now the condition is "NOT (button #0 released AND button #1
	released)". Or in other words: "button #0 pressed OR button #1
	pressed". Nice, isn't it?!

 */
	if (events & MU_BUTTON)
	{
#if PENDING_BUTTONS
		if (pending_button.client == client)
		{
			DIAG((D.mouse,-1,"pending_button multi %d\n", pending_button.state));
			pending_button.client = nil;			/* is single shot. */
			multi_intout(pb->intout, MU_BUTTON);
			if (is_bevent(pending_button.state, pb->intin[1], pb->intin[2], pb->intin[3], 1))
				return XAC_DONE;
		}
		else
#endif
		if (pb->intin[3] == 0 && (pb->intin[1] & 0x100) == 0)	/* real button up */
		{
			multi_intout(pb->intout, MU_BUTTON);		/* HR: we need the vq_mouse anyhow, might as well do the rest. */
			if (is_bevent(pb->intout[3], pb->intin[1], pb->intin[2], 0, 2))
				return XAC_DONE;
		}

		new_waiting_for |= XAWAIT_BUTTON;		/* Flag the app as waiting for button changes */
		ret = XAC_BLOCK;
	}

	if (events & MU_KEYBD)		
	{
		if (pending_key_strokes(pb, MU_KEYBD))			/* HR */
			return XAC_DONE;
		else
			new_waiting_for |= XAWAIT_KEY;			/* Flag the app as waiting for keypresses */
		ret = XAC_BLOCK;
	}

/* HR: event data are now in the client structure */

	if (events & (MU_M1|MU_M2))
	{		
		memset(&client->em,0,sizeof(XA_MOUSE_RECT));
		ret = XAC_BLOCK;
		if (events & MU_M1)					/* Mouse rectangle tracking */
		{
			new_waiting_for |= XAWAIT_MOUSE1;
			client->em.m1 = *((RECT *)&pb->intin[5]);
			client->em.flags = (long)(pb->intin[4]) | MU_M1;
		}

		if (events & MU_M2)
		{
			new_waiting_for |= XAWAIT_MOUSE2;
			client->em.m2 = *((RECT *)&pb->intin[10]);
			client->em.flags |= (long)((pb->intin[9]) << 1) | MU_M2;
		}
	}

	if (events & MU_MESAG)
	{
		if (pending_msgs(lock, client, pb, MU_MESAG))		/* HR */
			return XAC_DONE;

		new_waiting_for |= XAWAIT_MESSAGE;		/* Mark the client as waiting for messages */
	
		if (ret == XAC_DONE)
			ret = XAC_BLOCK;
	}
	
	/* HR: a zero timer (immediate timout) is catered for in the kernel. */

	if (events & MU_TIMER)
	{
		if (pb->intin[15])			/* high part */
			ret = 0xffff0000L | XAC_M_TIMEOUT;
		else if (pb->intin[14])
			ret = ((long)pb->intin[14] << 16) | XAC_M_TIMEOUT;
		
		DIAG((D.multi,-1,"Timer val: %d,%d ret: 0x%lx\n", pb->intin[14], pb->intin[15], ret));

		if (ret & XAC_M_TIMEOUT)
		{
			new_waiting_for |= XAWAIT_TIMER;	/* Flag the app as waiting for a timer */
		} elif (new_waiting_for == 0)			/* HR: Do not loose the other events */
		{
			multi_intout(pb->intout, MU_TIMER);	/* HR: fill out the mouse data */

		/* Is this the cause of loosing the key's at regular intervals? */
			new_waiting_for = 0;
DIAG((D.multi,-1,"Done timer for %d\n", client->pid));
			ret = XAC_DONE;
		} else
		{
			new_waiting_for |= XAWAIT_TIMER;	/* HR: make it 1 millisecond */
			ret = (1L << 16) | XAC_M_TIMEOUT;
		}
	}

	if (new_waiting_for)					/* If we actually recognised any of the codes, then set the multi flag */
		new_waiting_for |= XAWAIT_MULTI;
	
	client->waiting_for = new_waiting_for;			/* Flag the app as waiting */

	return ret;
}

/*
 *	Cancel an event_multi()
 *	- Called when any one of the events we were waiting for occurs
 */
void cancel_evnt_multi(XA_CLIENT *client)
{
	client->waiting_for = 0;
}

/*
 *	AES message handling
 */
AES_function XA_evnt_mesag	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
/* HR: static pid array */

	CONTROL(0,1,1)

	if (pending_msgs(lock, client, pb, 1))		/* HR */
		return XAC_DONE;
	
	client->waiting_for = XAWAIT_MESSAGE;	/* Mark the client as waiting for messages */
	client->waiting_pb = pb;

	return XAC_BLOCK;
}

/*
 *	evnt_button() routine
 */
AES_function XA_evnt_button	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short mb, x, y;

	CONTROL(3,5,1)

#if PENDING_BUTTONS
	if (pending_button.client == client)
	{
		DIAG((D.mouse,-1,"pending_button %d\n", pending_button.state));
		pending_button.client = nil;			/* is single shot. */
		multi_intout(pb->intout, MU_BUTTON);
		if (is_bevent(pending_button.state, pb->intin[0], pb->intin[1], pb->intin[2], 3))
			return XAC_DONE;
	}
	else
#endif
	if (pb->intin[2] == 0)
	{
		/* HR: multi_intout includes vq_mouse/keys */
		multi_intout(pb->intout, 0);		/* because this is 'button_up' no of clicks is 0. */
		if (is_bevent(pb->intout[3], pb->intin[0], pb->intin[1], 0, 4))
			return XAC_DONE;
	}

/* HR static pid array */

	client->waiting_for = XAWAIT_BUTTON;	/* Flag the app as waiting for messages */
	client->waiting_pb = pb;				/* Store a pointer to the AESPB to fill when the event */

	return XAC_BLOCK;	/* Returning FALSE blocks the client app to wait for the event */
}

/*
 *	evnt_keybd() routine
 */
AES_function XA_evnt_keybd	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(0,1,0)

	if (pending_key_strokes(pb, 0))			/* HR */
		return XAC_DONE;

	client->waiting_for = XAWAIT_KEY;	/* Flag the app as waiting for messages */
	client->waiting_pb = pb;			/* Store a pointer to the AESPB to fill when the event */
												/* finally arrives. */
	return XAC_BLOCK;	/* Returning FALSE blocks the client app to wait for the event */
}

/*
 *	Event Mouse
 */
AES_function XA_evnt_mouse	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(5,5,0)

	client->waiting_for = XAWAIT_MOUSE1;	/* Flag the app as waiting for mouse events */
	client->waiting_pb = pb;				/* Store a pointer to the AESPB to fill when the event */
										/*  finally arrives. */
	memset(&client->em,0,sizeof(XA_MOUSE_RECT));
	client->em.m1 = *((RECT *)&pb->intin[1]);
	client->em.flags = (long)(pb->intin[0]) | MU_M1;

	return XAC_BLOCK;	/* Returning FALSE blocks the client app to wait for the event */
}

/*
 *	Evnt_timer()
 */

AES_function XA_evnt_timer	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	union
	{
		unsigned long ul;
		unsigned short us[2];
	} ret;

	CONTROL(2,1,0)

	ret.us[1] = XAC_T_TIMEOUT;	
	if (pb->intin[1])
		ret.us[0] = 0xffff;
	else
		ret.us[0] = pb->intin[0];
		
	client->waiting_pb = pb;			/* Store a pointer to the AESPB to fill when the event */
												/*  finally arrives. */
	client->waiting_for = XAWAIT_TIMER;	/* Flag the app as waiting for a timer */

	return ret.ul;
}
