/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include "xa_types.h"
#include "xa_globl.h"
#include "objects.h"
#include "xa_graf.h"		/* for graf_mouse */

#define THICKNESS	2

/*
 *	This routine behaves a lot like graf_watchbox() - in fact, XaAES's
 *	graf_watchbox calls this.
 *	
 *	The differance here is that we handle colour icons properly (something
 *	that AES4.1 didn't do).
 *	
 *	I've re-used this bit from the WINDIAL module I wrote for the DULIB GEM library.
 *	
 *	This routine assumes that any clipping is going to be done elsewhere
 *	before we get to here.
 */
short watch_object(OBJECT *dial, short ob, short in_state, short out_state)
{
	OBJECT *the_object = dial + ob;
	short pobf = -2, obf = ob, mx, my, mb, x, y, /* w, h, */ omx, omy;
	
	vq_mouse(V_handle, &mb, &omx, &omy);
	
	object_abs_coords(dial, ob, &x, &y);	/* call XaAES's internal objc_offset() */
/*	w = the_object->r.w + 2;
	h = the_object->r.h + 2;
*/	x--;
	y--;

	if (!mb)	/* If mouse button is already released, assume that was just a click, so select */
	{
		(dial + ob)->ob_state = in_state;
		v_hide_c(V_handle);
		display_object(dial, ob,x - the_object->r.x + 1, y - the_object->r.y + 1);
		v_show_c(V_handle, 1);
	} else
	{
	
		while (mb)	/* This loop contains a pretty busy wait, but I don't think it's to */
		{			/* much of a problem as the user is interacting with it continuously. */
			vq_mouse(V_handle, &mb, &mx, &my);
		
			if ((mx != omx) || (my != omy))
			{
				omx = mx;
				omy = my;
				obf = find_object(dial, ob, 10, mx, my);
		
				if (obf == ob)
					(dial + ob)->ob_state = in_state;
				else
					(dial + ob)->ob_state = out_state;
				
				if (pobf != obf)
				{
					pobf = obf;
					v_hide_c(V_handle);		
					display_object(dial, ob, x - the_object->r.x + 1, y - the_object->r.y + 1);
					v_show_c(V_handle, 1);
				}
			}
		}
		
	}

	vsf_interior(V_handle, FIS_SOLID);
	vswr_mode(V_handle, MD_TRANS);

	if (obf == ob)
		return 1;
	else
		return 0;
}

static
void draw_box(short x, short y, short w, short h, short thickness)
{
	short pnt[4];

	if (thickness > 0)
	{
		pnt[0] = x;
		pnt[1] = y;
		pnt[2] = x + w - 1;
		pnt[3] = y + (thickness - 1);
		v_bar(V_handle, pnt);

		pnt[1] = y + h - 1 - (thickness - 1);
		pnt[3] = y + h - 1;
		v_bar(V_handle, pnt);

		pnt[1] = y + thickness;
		pnt[2] = x + (thickness - 1);
		pnt[3] = y + h - 2 - (thickness - 1);
		v_bar(V_handle, pnt);

		pnt[0] = x + w - 1 - (thickness - 1);
		pnt[2] = x + w - 1;
		v_bar(V_handle, pnt);
	}
	else
	if (thickness < 0)
	{
		thickness = -thickness;
		pnt[0] = x - thickness;
		pnt[1] = y - thickness;
		pnt[2] = x + w - 1 + thickness;
		pnt[3] = y - 1;
		v_bar(V_handle, pnt);

		pnt[1] = y + h;
		pnt[3] = y + h + (thickness - 1);
		v_bar(V_handle, pnt);

		pnt[1] = y - thickness;
		pnt[2] = x - 1;
		pnt[3] = y + h - 1;
		v_bar(V_handle, pnt);

		pnt[0] = x + w;
		pnt[2] = x + w + (thickness - 1);
		v_bar(V_handle, pnt);
	}
}

/*
 *	Ghost outline dragging routines - all coded to work with full rectangles instead of outlines
 */

void drag_box(short w, short h, short xs, short ys, short bx, short by, short bw, short bh, short *xe, short *ye)
{
	short x, y, mb, idx, idy, dx, dy, odx = -1, ody = -1;
	short pnt[4];
	
	vq_mouse(V_handle, &mb, &x, &y);
	vswr_mode(V_handle, MD_XOR);

	vsf_color(V_handle, BLACK);
	vsf_perimeter(V_handle, 0);

	clear_clip();
	
	idx = x - xs;
	idy = y - ys;
	
	while(mb)
	{
		vq_mouse(V_handle, &mb, &x, &y);

		dx = x - idx;
		dy = y - idy;
		
		if (dx < bx)		/* Ensure we are inside the bounding rectangle */
			dx = bx;
		else
		if (dx + w > bx + bw)
			dx = bx + bw - w - 1;

		if (dy < by)
			dy = by;
		else
		if (dy + h > by + bh)
			dy = by + bh - h - 1;
		
		if ((dx != odx) || (dy != ody))
		{
			v_hide_c(V_handle);
			if (odx != -1)
#if SOLID_BOXES
				v_bar(V_handle, pnt);		/* This is a real man's GUI - dragged boxes can be done as solid */

			pnt[0] = dx;
			pnt[1] = dy;
			pnt[2] = dx + w;
			pnt[3] = dy + h;
			v_bar(V_handle, pnt);
#else
			draw_box(odx, ody, w, h, THICKNESS);
			draw_box(dx, dy, w, h, THICKNESS);				
#endif
			v_show_c(V_handle, 1);
			odx = dx;
			ody = dy;
		}
	}
	
	if (odx != -1)
	{
		v_hide_c(V_handle);
		draw_box(dx, dy, w, h, THICKNESS);
		v_show_c(V_handle, 1);
	}

	vswr_mode(V_handle, MD_TRANS);
	
	*xe = dx;
	*ye = dy;
}

/* HR: dx, dy for use with sizer: If you click in the middle of the sizer,
       than change your mind and release
       the button, without moving the mouse,
       the window size will not change.
*/
void rubber_box(short sx, short sy, short minw, short minh, short dx, short dy, short *lastw, short *lasth)
{
	short x, y, mb, dw, dh, odw = -1, odh = -1;
	short pnt[4];
	
	vswr_mode(V_handle, MD_XOR);		/* HR: !!!!! that's why perimeter works !!!!! */
	vq_mouse(V_handle, &mb, &x, &y);

	vsf_color(V_handle, BLACK);
	vsf_perimeter(V_handle, 0);

	while(mb)
	{
		vq_mouse(V_handle, &mb, &x, &y);
		dw = x - sx;
		dh = y - sy;
		
		if (dw < minw)
			dw = minw;

		if (dh < minh)
			dh = minh;
		
		if ((dw != odw) || (dh != odh))
		{
			v_hide_c(V_handle);
			if (odw != -1)
				draw_box(sx, sy, odw + dx, odh + dy, THICKNESS);
			draw_box(sx, sy, dw + dx, dh + dy, THICKNESS);
			v_show_c(V_handle, 1);
			odw = dw;
			odh = dh;
		}
	}
	
	if (odw != -1)
	{
		v_hide_c(V_handle);
		draw_box(sx, sy, dw + dx, dh + dy, THICKNESS);
		v_show_c(V_handle, 1);
	}

	vswr_mode(V_handle, MD_TRANS);
	
	*lastw = dw + dx;
	*lasth = dh + dy;
}

/*
 *	INTERFACE TO INTERACTIVE BOX ROUTINES
 */

AES_function XA_graf_dragbox	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short x, y;

	CONTROL(8,3,0)

	drag_box(pb->intin[0],
				pb->intin[1],
				pb->intin[2],
				pb->intin[3],
				pb->intin[4],
				pb->intin[5],
				pb->intin[6],
				pb->intin[7],
				&x, &y);
	
	pb->intout[0] = 1;
	pb->intout[1] = x;
	pb->intout[2] = y;

DIAG((D.graf,client->pid,"_drag_box\n"));
	return XAC_DONE;
}

AES_function XA_graf_rubberbox	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short w, h;
	
	CONTROL(4,3,0)

	rubber_box(pb->intin[0],
				pb->intin[1],
				pb->intin[2],
				pb->intin[3],
				0, 0,
				&w, &h);

	pb->intout[0] = 1;
	pb->intout[1] = w;
	pb->intout[2] = h;
DIAG((D.graf,client->pid,"_rubbox x=%d, y=%d, w=%d, h=%d\n",pb->intin[0],pb->intin[1],w,h));	
	return XAC_DONE;
}

AES_function XA_graf_watchbox	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(4,1,1)

	pb->intout[0] = watch_object((OBJECT *)pb->addrin[0],
											pb->intin[1],
											pb->intin[2],
											pb->intin[3]);
DIAG((D.graf,client->pid,"_watchbox\n"));
	return XAC_DONE;
}

/* If shrinkbox isnt there, why then growbox? 
AES_function XA_graf_growbox		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(8,1,0)

	short x = pb->intin[0];
	short y = pb->intin[1];
	short w = pb->intin[2];
	short h = pb->intin[3];
	short xe = pb->intin[4];
	short ye = pb->intin[5];
	short we = pb->intin[6];
	short he = pb->intin[7];
	short dx, dy, dw, dh;
	short f;
	
	dx = (xe - x) / GRAF_STEPS;
	dy = (ye - y) / GRAF_STEPS;
	dw = (we - w) / GRAF_STEPS;
	dh = (he - h) / GRAF_STEPS;

	vswr_mode(V_handle, MD_XOR);

	v_hide_c(V_handle);

	for (f = 0; f < GRAF_STEPS; f++)
	{
		draw_2d_box(x, y, w, h, 1, BLACK);	/* Draw initial growing outline */
		x += dx;
		y += dy;
		w += dw;
		h += dh;
#if 0
		if (f % 2)
			Vsync();
#endif
	}
	
	x = pb->intin[0];						/* Reset to initial area */
	y = pb->intin[1];
	w = pb->intin[2];
	h = pb->intin[3];

	for (f = 0; f < GRAF_STEPS; f++)	/* Erase growing outline */
	{
		draw_2d_box(x, y, w, h, 1, BLACK);
		x += dx;
		y += dy;
		w += dw;
		h += dh;
#if 0
		if (f % 2)
			Vsync();
#endif
	}

	v_show_c(V_handle, 1);

	vswr_mode(V_handle, MD_TRANS);
	
	pb->intout[0] = 1;
	
DIAG((D.graf,client->pid,"_growbox\n"));
	return XAC_DONE;
}
*/

AES_function XA_graf_movebox		 /*(LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short w = pb->intin[0];
	short h = pb->intin[1];
	short x = pb->intin[2];
	short y = pb->intin[3];
	short xe = pb->intin[4];
	short ye = pb->intin[5];
	short dx, dy;
	short f;
	
	CONTROL(6,1,0)

	dx = (xe - x) / GRAF_STEPS;
	dy = (ye - y) / GRAF_STEPS;

	vswr_mode(V_handle, MD_XOR);

	v_hide_c(V_handle);

	for (f = 0; f < GRAF_STEPS; f++)	/* Draw initial images */
	{
		draw_2d_box(x, y, w, h, 1, BLACK);
		x += dx;
		y += dy;
#if 0
		if (f % 2)
			Vsync();
#endif
	}
	
	x = pb->intin[2];					/* Reset to go back over same area */
	y = pb->intin[3];

	for (f = 0; f < GRAF_STEPS; f++)	/* Erase them again */
	{
		draw_2d_box(x, y, w, h, 1, BLACK);
		x += dx;
		y += dy;
#if 0
		if (f % 2)
			Vsync();
#endif
	}

	v_show_c(V_handle, 1);

	vswr_mode(V_handle, MD_TRANS);
	
	pb->intout[0] = 1;
	
DIAG((D.graf,client->pid,"_movebox\n"));
	return XAC_DONE;
}

/*
 * Mouse definition block - standard GEM arrow
 */
static
MFORM M_ARROW_MOUSE =
{	0x0000, 0x0000, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0xC000, 0xE000, 0xF000, 0xF800,
	0xFC00, 0xFE00, 0xFF00, 0xFF80,
	0xFFC0, 0xFFE0, 0xFE00, 0xEF00,
	0xCF00, 0x8780, 0x0780, 0x0380,
/* Cursor data */
	0x0000, 0x4000, 0x6000, 0x7000,
	0x7800, 0x7C00, 0x7E00, 0x7F00,
	0x7F80, 0x7C00, 0x6C00, 0x4600,
	0x0600, 0x0300, 0x0300, 0x0000
},

/*
 * Mouse definition block - standard GEM busy bee
 */

M_BEE_MOUSE =
{	0x0008, 0x0008, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x1C7E, 0x1CFF, 0x1CFF, 0xEFFF,
	0xFFFF, 0xFFFF, 0x3FFE, 0x3FFC,
	0x7FFE, 0xFFFE, 0xFFFF, 0xFFFF,
	0xFFFF, 0xFFFF, 0xFEFF, 0x7C3E,
/* Cursor data */
	0x0800, 0x083C, 0x0062, 0x06C2,
	0xC684, 0x198A, 0x1B54, 0x06E0,
	0x1D58, 0x33FC, 0x6160, 0x42DE,
	0x44D8, 0x4A56, 0x3414, 0x0000
},

/*
 * Mouse definition block standard GEM open hand
 */
#if 0
M_HAND_MOUSE =
{	0x0008, 0x0008, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x0300, 0x1FB0, 0x3FF8, 0x3FFC,
	0x7FFE, 0xFFFE, 0xFFFE, 0x7FFF,
	0x7FFF, 0xFFFF, 0xFFFF, 0x7FFF,
	0x3FFF, 0x0FFF, 0x01FF, 0x003F,
/* Cursor data */
	0x0300, 0x1CB0, 0x2448, 0x2224,
	0x7112, 0x9882, 0x8402, 0x4201,
	0x7001, 0x9801, 0x8401, 0x4000,
	0x3000, 0x0E00, 0x01C0, 0x0030
},
#else
/* non standard sliding hand */ 
M_HAND_MOUSE =
{	4,4,1,0,1,
 {	0x7A00, 0x7F40, 0xFFE0, 0xFFF0, 
	0xFFF8, 0x3FFC, 0x0FFE, 0xC7FF, 
	0xEFFF, 0xFFFF, 0xFFFF, 0x7FFF, 
	0x3FFF, 0x0FFF, 0x03FE, 0x00FC },
 {	0x5000, 0x2A00, 0x1540, 0xCAA0, 
	0x3550, 0x0AA8, 0x0404, 0x0252, 
	0xC402, 0xAA01, 0x5001, 0x3001, 
	0x0C01, 0x0300, 0x00C0, 0x0000 }
},
#endif

/*
 * Mouse definition block - standard GEM outlined cross
 */

M_OCRS_MOUSE =
{	0x0007, 0x0007, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x07C0, 0x07C0, 0x06C0, 0x06C0,
	0x06C0, 0xFEFE, 0xFEFE, 0xC006,
	0xFEFE, 0xFEFE, 0x06C0, 0x06C0,
	0x06C0, 0x07C0, 0x07C0, 0x0000,
/* Cursor data */
	0x0000, 0x0380, 0x0280, 0x0280,
	0x0280, 0x0280, 0x7EFC, 0x4004,
	0x7EFC, 0x0280, 0x0280, 0x0280,
	0x0280, 0x0380, 0x0000, 0x0000
},

/*
 * Mouse definition block - standard GEM pointing finger
 */

M_POINT_MOUSE =
{	0x0000, 0x0000, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x3000, 0x7C00, 0x7E00, 0x1F80,
	0x0FC0, 0x3FF8, 0x3FFC, 0x7FFC,
	0xFFFE, 0xFFFE, 0x7FFF, 0x3FFF,
	0x1FFF, 0x0FFF, 0x03FF, 0x00FF,
/* Cursor data */
	0x3000, 0x4C00, 0x6200, 0x1980,
	0x0C40, 0x32F8, 0x2904, 0x6624,
	0x93C2, 0xCF42, 0x7C43, 0x2021,
	0x1001, 0x0C41, 0x0380, 0x00C0
},

/*
 * Mouse definition block - standard GEM thin cross
 */

M_TCRS_MOUSE =
{	0x0007, 0x0007, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x0380, 0x0380, 0x0380, 0x0380,
	0x0280, 0x0280, 0xFEFE, 0xF01E,
	0xFEFE, 0x0280, 0x0280, 0x0380,
	0x0380, 0x0380, 0x0380, 0x0000,
/* Cursor data */
	0x0000, 0x0100, 0x0100, 0x0100,
	0x0100, 0x0100, 0x0100, 0x7FFC,
	0x0100, 0x0100, 0x0100, 0x0100,
	0x0100, 0x0100, 0x0000, 0x0000
},

/*
 * Mouse definition block - standard GEM thick cross
 */

M_THKCRS_MOUSE =
{	0x0007, 0x0007, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x07C0, 0x07C0, 0x07C0, 0x07C0,
	0x07C0, 0xFFFE, 0xFFFE, 0xFFFE,
	0xFFFE, 0xFFFE, 0x07C0, 0x07C0,
	0x07C0, 0x07C0, 0x07C0, 0x0000,
/* Cursor data */
	0x0000, 0x0380, 0x0380, 0x0380,
	0x0380, 0x0380, 0x7FFC, 0x7FFC,
	0x7FFC, 0x0380, 0x0380, 0x0380,
	0x0380, 0x0380, 0x0000, 0x0000
},

/*
 * Mouse definition block - standard GEM text cursor
 */

M_TXT_MOUSE =
{	0x0007, 0x0007, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x7E7E, 0x7FFE, 0x07E0, 0x03C0,
	0x03C0, 0x03C0, 0x03C0, 0x03C0,
	0x03C0, 0x03C0, 0x03C0, 0x03C0,
	0x03C0, 0x07E0, 0x7FFE, 0x7E7E,
/* Cursor data */
	0x3C3C, 0x0660, 0x03C0, 0x0180,
	0x0180, 0x0180, 0x0180, 0x0180,
	0x0180, 0x0180, 0x0180, 0x0180,
	0x0180, 0x03C0, 0x0660, 0x3C3C
},

/*
 * Mouse definition block - XaAES Data Uncertain logo cursor
 */

M_BUBD_MOUSE =
{	0x0007, 0x0008, 0x0002, 0x0000, 0x0001,
/* Mask data */
	0x07C0, 0x1FF0, 0x3FF8, 0x7FFC,
	0x7FFC, 0xFFFE, 0xFFFE, 0xFFFE,
	0xFFFE, 0xFFFE, 0x7FFC, 0x7FFC,
	0x3FF8, 0x1FF0, 0x07C0, 0x0000,
/* Cursor data */
	0x07C0, 0x1930, 0x2AA8, 0x5094,
	0x7114, 0x9012, 0x9112, 0x9FF2,
	0x8002, 0x8FE2, 0x4A24, 0x4B24,
	0x2B28, 0x1A30, 0x07C0, 0x0000
},

/*
 * Mouse definition block - XaAES re-sizer cursor
 */

M_SIZER_MOUSE =
{	0x0007, 0x0008, 0x0002, 0x0000, 0x0001,
/* Mask data */
	0x0000, 0xFE00, 0x8100, 0x8E00,
	0x8400, 0xA200, 0xB100, 0xA888,
	0x4454, 0x0234, 0x0114, 0x0084,
	0x01C4, 0x0204, 0x01FC, 0x0000,
/* Cursor data */
	0x0000, 0x0000, 0x7E00, 0x7000,
	0x7800, 0x5C00, 0x4E00, 0x4700,
	0x0388, 0x01C8, 0x00E8, 0x0078,
	0x0038, 0x01F8, 0x0000, 0x0000
},

/*
 * Mouse definition block - XaAES mover cursor
 */

M_MOVER_MOUSE =
{	0x0008, 0x0007, 0x0001, 0x0000, 0x0001,
/* Mask data */
	0x0000, 0x0080, 0x01C0, 0x0080,
	0x0080, 0x0080, 0x1084, 0x3FFE,
	0x1084, 0x0080, 0x0080, 0x0080,
	0x01C0, 0x0080, 0x0000, 0x0000,
/* Cursor data */
	0x0080, 0x0140, 0x0220, 0x0140,
	0x0140, 0x1144, 0x2F7A, 0x4001,
	0x2F7A, 0x1144, 0x0140, 0x0140,
	0x0220, 0x0140, 0x0080, 0x0000
};

/*
 *	AES graf_mouse() routines
 *	Small extension to give a couple of extra cursor shapes
 *	(Data Uncertain logo, mover & resizer)
 */

void graf_mouse(short m_shape, MFORM *mf)
{
	switch(m_shape)
	{
	case M_ON:
		v_show_c(V_handle, 1);
		return;
	case M_OFF:
		v_hide_c(V_handle);
		return;
	case ARROW:
		vsc_form(V_handle, &M_ARROW_MOUSE);
		break;
	case TEXT_CRSR:
		vsc_form(V_handle, &M_TXT_MOUSE);
		break;
	case HOURGLASS:
		vsc_form(V_handle, &M_BEE_MOUSE);
		break;
	case POINT_HAND:
		vsc_form(V_handle, &M_POINT_MOUSE);
		break;
	case FLAT_HAND:
		vsc_form(V_handle, &M_HAND_MOUSE);
		break;
	case THIN_CROSS:
		vsc_form(V_handle, &M_TCRS_MOUSE);
		break;
	case THICK_CROSS:
		vsc_form(V_handle, &M_THKCRS_MOUSE);
		break;
	case OUTLN_CROSS:
		vsc_form(V_handle, &M_OCRS_MOUSE);
		break;
	case USER_DEF:
		if (!mf)
			mf = &M_BUBD_MOUSE;	/* HR: Scare people ;-) */
		vsc_form(V_handle, mf);
		break;
	case XACRS_BUBBLE_DISC:		/* The Data Uncertain logo */
		vsc_form(V_handle, &M_BUBD_MOUSE);
		break;
	case XACRS_RESIZER:			/* The 'resize window' cursor */
		vsc_form(V_handle, &M_SIZER_MOUSE);
		break;
	case XACRS_MOVER:			/* The 'move window' cursor */
		vsc_form(V_handle, &M_MOVER_MOUSE);
		break;
	}
	v_hide_c(V_handle);				/* Hide/reveal cursor to update on screen image immediately */
	v_show_c(V_handle, 1);
}

/* Slight differance from GEM here - each application can have a different mouse form, */
/* and the one that is used is for the client with the top window. */
/* A non-top application can still hide & show the mouse though, to ensure that redraws */
/* are done correctly */
AES_function XA_graf_mouse	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short m = pb->intin[0];
	
	CONTROL(1,1,1)

/* Store the mouse shape requested if this isn't a turn on/off request. */
	if ((m != M_OFF) && (m != M_ON))
	{
		client->client_mouse = m;
		client->client_mouse_form = pb->addrin[0];	
/* Only actually set the cursor shape if the top window is our client's  */
/* or there are no windows open at all */
		if ((window_list == root_window) || (window_list->owner == client->pid))
			graf_mouse(m, pb->addrin[0]);
	} else
	{
		graf_mouse(m, NULL);	/* Any client can hide the mouse (required for redraws by clients */
								/* that aren't top) */
	}

/* Always return no error */
	pb->intout[0] = 1;

DIAG((D.graf,client->pid,"_mouse\n"));
	return XAC_DONE;
}

AES_function XA_graf_handle	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(0,5,0)

	pb->intout[0] = P_handle;		/* graf_handle returns the physical workstation handle */
	pb->intout[1] = display.c_max_w;
	pb->intout[2] = display.c_max_h;
	pb->intout[3] = display.c_max_w+2;
	pb->intout[4] = display.c_max_h+2;
	
DIAG((D.graf,client->pid,"_handle\n"));
	return XAC_DONE;
}

AES_function XA_graf_mkstate	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short x, y, mb, k;
	
	CONTROL(0,5,0)

	vq_mouse(V_handle, &mb, &x, &y);
	vq_key_s(V_handle, &k);
	pb->intout[0] = 1;
	pb->intout[1] = x;
	pb->intout[2] = y;
	pb->intout[3] = mb;
	pb->intout[4] = k;

DIAG((D.graf,client->pid,"_mkstate: %d/%d, m=0x%x, k=0x%x\n",x,y,mb,k));
	return XAC_DONE;		/* HR: was TRUE */
}
