/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <stdlib.h>
#include <osbind.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifdef LATTICE
#undef abs		/* MiNTlib (PL46) #define is buggy! */
#define abs(i)	__builtin_abs(i)
#endif

#include "xa_types.h"
#include "xa_globl.h"

#include "rectlist.h"
#include "objects.h"
#include "menuwidg.h"

/*
 *	OBJECT TREE ROUTINES
 *	- New version of the object display routine modularises the whole system.
 */
#if GENERATE_DIAGS

short global_pid_for_draw = 0;

static char *ob_types[] =
{
	"box",
	"text",
	"boxtext",
	"image",
	"progdef",
	"ibox",
	"button",
	"boxchar",
	"string",
	"ftext",
	"fboxtext",
	"icon",
	"title",
	"cicon",
	"xaaes slist",
	"popup",
	"resv",
	"shortcut",
	"39",
	"40"
};

char other[80],nother[160];

char *object_type(OBJECT *tree, short t)
{
	unsigned short ty = tree[t].ob_type,tx;
	if (ty >= G_BOX && ty < G_MAX)
		return ob_types[ty-G_BOX];
	tx = ty&0xff;
	if (tx >= G_BOX && tx < G_MAX)
		sprintf(other,"ext: 0x%x + %s",ty>>8,ob_types[tx-G_BOX]);
	else
		sprintf(other,"unknown: 0x%x,%d",ty,ty);
	return other;
}

char *object_txt(OBJECT *tree, short t)			/* HR: I want to know the culprit in a glance */
{
	short ty = tree[t].ob_type;
	*nother = 0;
	switch (ty&0xff)
	{
		case G_FTEXT:
		case G_TEXT:
		case G_BOXTEXT:
		case G_FBOXTEXT:
		{
			TEDINFO *ted = (TEDINFO *)ob_spec(tree + t);
			sprintf(nother," '%s'",ted->te_ptext);
			break;
		}
		case G_BUTTON:
		case G_TITLE:
		case G_STRING:
		case G_SHORTCUT:
			sprintf(nother," '%s'",(char *)ob_spec(tree + t));
	}	
	return nother;
}

#endif

ObjectDisplay *objc_jump_table[G_MAX];

short global_clip[4];	/* Bloody progdefs need to know the clip rectangle */

void set_clip(const RECT *r)
{
	rtopxy(global_clip,r);
	vs_clip(V_handle, 1, global_clip);
}

/*
 * HR change ob_spec
 */
void set_obspec(OBJECT *root, short s_ob, unsigned long cl)
{
#if 0
	root[s_ob].ob_spec = (void *)cl;
#else
 #ifdef do_not_do_this		/* __GNUC__ */

/* HR: Choice statement as Lvalue: bad luck!
	Try Algol68 (32 years ago this would be OK) :-P
	but we live in postmodern times, do we? */

	ob_spec(&root[s_ob]) = (void *)cl;		
 #else
	if (root[s_ob].ob_flags & 0x100)
		*((void **)root[s_ob].ob_spec) = (void *)cl;
	else
		root[s_ob].ob_spec = (void *)cl;
 #endif
#endif
}

/*
 * Set clipping to entire screen
 */
void clear_clip(void)
{
	rtopxy(global_clip,&display.r);
	vs_clip(V_handle, 1, global_clip);
}

#if 1

/* HR: I wonder which one is faster;
		This one is smaller. and easier to follow. */
void g2d_box(short b, RECT *r, short colour)
{
	if (b)
	{
		short j = b/-abs(b);	/* stepper with opposite sign */
		if (b >  4) b =  4;
		if (b < -4) b = -4;
		vsl_color(V_handle, colour);
		while (b)
			gbox(-b, r), b+=j;
	}
}

/* sheduled for redundancy */
void draw_2d_box(short x, short y, short w, short h, short border_thick, short colour)
{
	RECT r;
	r.x = x, r.y = y, r.w = w, r.h = h;
	g2d_box(border_thick, &r, colour);
}

#else

/*
 * Draw a 2d box outline (allows 'proper' thickness - 1, 2, 3, etc)
 * Note: expects writing mode to be set up by caller.
 */
void draw_2d_box(short x, short y, short w, short h, short border_thick, short colour)
{
	short coords[4];

	vsf_interior(V_handle, FIS_SOLID);
	vsf_color(V_handle, colour);
			
	if (border_thick > 0)
	{
		coords[0] = x;
		coords[1] = y;
		coords[2] = x + w - 1;
		coords[3] = y + (border_thick - 1);
		v_bar(V_handle, coords);

		coords[1] = y + h - 1 - (border_thick - 1);
		coords[3] = y + h - 1;
		v_bar(V_handle, coords);

		coords[1] = y + border_thick;
		coords[2] = x + (border_thick - 1);
		coords[3] = y + h - 2 - (border_thick - 1);
		v_bar(V_handle, coords);

		coords[0] = x + w - 1 - (border_thick - 1);
		coords[2] = x + w - 1;
		v_bar(V_handle, coords);
	} else if (border_thick < 0)
	{
		border_thick = -border_thick;
		coords[0] = x - border_thick;
		coords[1] = y - border_thick;
		coords[2] = x + w - 1 + border_thick;
		coords[3] = y - 1;
		v_bar(V_handle, coords);

		coords[1] = y + h;
		coords[3] = y + h + (border_thick - 1);
		v_bar(V_handle, coords);

		coords[1] = y - border_thick;
		coords[2] = x - 1;
		coords[3] = y + h - 1;
		v_bar(V_handle, coords);

		coords[0] = x + w;
		coords[2] = x + w + (border_thick - 1);
		v_bar(V_handle, coords);
	}
}

void g2d_box(short b, RECT *r, short colour)
{
	draw_2d_box(r.x, r.y, r.w, r.h, sb, colour);
}


#endif

/*
 * Initialise the object display jump table
 */
void init_objects(void)
{
	short f;

	for (f = 0; f < G_MAX; f++)
		objc_jump_table[f] = NULL;			/* Anything with a NULL pointer won't get called */
	
	objc_jump_table[G_BOX] = d_g_box;
	objc_jump_table[G_TEXT] = d_g_text;
	objc_jump_table[G_BOXTEXT] = d_g_boxtext;
	objc_jump_table[G_IMAGE] = d_g_image;
	objc_jump_table[G_PROGDEF] = d_g_progdef;
	objc_jump_table[G_IBOX] = d_g_ibox;
	objc_jump_table[G_BUTTON] = d_g_button;
	objc_jump_table[G_BOXCHAR] = d_g_boxchar;
	objc_jump_table[G_STRING] = d_g_string;
	objc_jump_table[G_FTEXT] = d_g_ftext;
	objc_jump_table[G_FBOXTEXT] = d_g_fboxtext;
	objc_jump_table[G_ICON] = d_g_icon;
	objc_jump_table[G_TITLE] = d_g_title;
	objc_jump_table[G_CICON] = d_g_cicon;
	objc_jump_table[G_SLIST] = d_g_slist;
	objc_jump_table[G_SHORTCUT] = d_g_string;		/* HR */
}

/*
 * Display a primitive object
 */
#if UNCLEAN
#include "obsolete/disp_obj.h"
#else
void display_object(OBJECT *tree, short object, short parent_x, short parent_y)
{
	RECT r;
	OBJECT *ob = tree + object;
	ObjectDisplay *display_routine = NULL;		/* HR */
	ODC_PARM odc_p;
	short coords[10];
	/* HR: state_mask is for G_PROGDEF originally.
	   But it means that other objects must unflag what they
	   can do themselves in the same manner.
	   The best thing (no confusion) is to generalize the concept.
	   Which I did. :-) */
	short state_mask = (SELECTED|CROSSED|CHECKED|DISABLED|OUTLINED);
	short t = ob->ob_type & 0xff;

	r.x = parent_x + ob->r.x;
	r.y = parent_y + ob->r.y;
	r.w = ob->r.w; 
	r.h = ob->r.h;

	if (   r.x       > global_clip[2]	/* Exit immediately if object is totally outside clip area */
		|| r.x+r.w-1 < global_clip[0]
		|| r.y       > global_clip[3]
		|| r.y+r.h-1 < global_clip[1]
		)
		return;

	if (t < G_MAX)						/* HR */
		display_routine = objc_jump_table[t];	/* Get display routine for this type of object from jump table */

	if (display_routine == NULL)			/* If we don't have a display routine for a given object type, draw a box instead */
#if 0
		display_routine = objc_jump_table[G_IBOX];
#else
	{
		short i;
		DIAG((D.objc,global_pid_for_draw,"ob_type: %d(0x%x)\n", t, ob->ob_type));
/*		for (i=0; i<G_MAX; i++)
			DIAG((D.objc,global_pid_for_draw,"%08lx\n",objc_jump_table[i]));
*/		return;				/* HR: dont attempt doing what could be indeterminate!!! */
	}
#endif
		
	odc_p.tree = tree;					/* Fill in the object display parameter structure */
	odc_p.object = object;					
	odc_p.parent_x = parent_x;
	odc_p.parent_y = parent_y;
	odc_p.r = r;				/* HR: absolute RECT, ready for use everywhere. */
	odc_p.state_mask = &state_mask;

	vswr_mode(V_handle, MD_TRANS);		/* Better do this before AND after (fail safe) */

	(*display_routine)(&odc_p);			/* Call the appropriate display routine */

	vswr_mode(V_handle, MD_TRANS);

	/* Handle CROSSED object state: */
	if ((ob->ob_state & state_mask) & CROSSED)	/* HR */
	{
		vsl_color(V_handle, BLACK);
		coords[0] = r.x;
		coords[1] = r.y;
		coords[2] = r.x + r.w - 1;
		coords[3] = r.y + r.h - 1;
		v_pline(V_handle, 2, coords);
		coords[0] = r.x + r.w - 1;
		coords[2] = r.x;
		v_pline(V_handle, 2, coords);
	}

	/* Handle CHECKED object state: */
	if ((ob->ob_state & state_mask) & CHECKED)		/* HR */
	{
		vst_color(V_handle, BLACK);
		v_gtext(V_handle, r.x + 2, r.y, "\10");	/* ASCII 8 = checkmark */
	}

	/* Handle DISABLED state: */
	/* (May not look too hot in colour mode, but it's better than
		no disabling at all...) */
	if ((ob->ob_state & state_mask) & DISABLED)
	{
		static short pattern[16] =
		{
			0x5555, 0xaaaa, 0x5555, 0xaaaa,
			0x5555, 0xaaaa, 0x5555, 0xaaaa,
			0x5555, 0xaaaa, 0x5555, 0xaaaa,
			0x5555, 0xaaaa, 0x5555, 0xaaaa
		};

		vsf_color(V_handle, WHITE);
		vsf_udpat(V_handle, pattern, 1);
		vsf_interior(V_handle, FIS_USER);
		gbar(0, &r);
	}

	if ((ob->ob_state & state_mask) & SELECTED)
		write_selection(&r);

	vswr_mode(V_handle, MD_TRANS);
}
#endif

/*
 * Walk an object tree, calling display for each object
 */

short draw_object_tree(OBJECT *tree, short object, short depth)
{
	short next;
	short current = 0, rel_depth = 1, head;
	short x = 0, y = 0, start_drawing = FALSE;

	DIAG((D.objc,global_pid_for_draw,"draw_object_tree@%lx, start:%d depth:%d\n", tree, object, depth));

	depth++;
	
	do
	{

		if (current == object)
		{
			start_drawing = TRUE;
			rel_depth = 0;
		}

		if (start_drawing != 0)
		{
#if GENERATE_DIAGS
			{
				char *hide = (tree[current].ob_flags & HIDETREE) == 0 ? "" : " H";
				DIAG((D.o,global_pid_for_draw,"ob=%d,r=%d,s=%d%s; %s%s fl=0x%x, st=0x%x\n",
					 current, rel_depth, start_drawing, hide, object_type(tree,current),object_txt(tree,current),
					 tree[current].ob_flags,
					 tree[current].ob_state));
			}
#endif
			if ((tree[current].ob_flags & HIDETREE) == 0)
				display_object(tree, current, x, y);	/* Display this object */
		}

		head = tree[current].ob_head;
										/* Any non-hidden children? */
		if (   head != -1
			&& (tree[current].ob_flags & HIDETREE) == 0
			&& (
				   start_drawing == 0
				|| (   start_drawing != 0
				    && rel_depth<depth
				   )
			   )
			)
		{
			x += tree[current].r.x;
			y += tree[current].r.y;
			rel_depth++;
			current = head;
		} else
		{
			next = tree[current].ob_next;		/* Try for a sibling */
	
			while((next != -1)				/* Trace back up tree if no more siblings */
				&& (tree[next].ob_tail == current))
			{
				current = next;
				x -= tree[current].r.x;
				y -= tree[current].r.y;
				next = tree[current].ob_next;
				rel_depth--;
			}
			current = next;
		}
	
	} while (current != -1 && !(start_drawing && rel_depth < 1));

	vswr_mode(V_handle, MD_TRANS);
	vsf_interior(V_handle, FIS_SOLID);

#if GENERATE_DIAGS
	global_pid_for_draw = 0;
#endif
	return TRUE;
}

/*
 * Get the true screen coords of an object
 */

short object_abs_coords(OBJECT *tree, short object, short *obx, short *oby)
{
	short next;
	short current = 0;
	short x = 0, y = 0;
	
	do
	{
		if (current == object)	/* Found the object in the tree? cool, return the coords */
		{
			*obx = x + tree[current].r.x;
			*oby = y + tree[current].r.y;
			return 1;
		}

		if (tree[current].ob_head != -1)	/* Any children? */
		{
			x += tree[current].r.x;
			y += tree[current].r.y;
			current = tree[current].ob_head;
		} else
		{
			next = tree[current].ob_next;		/* Try for a sibling */

			while((next != -1) && (tree[next].ob_tail == current))
			{	/* Trace back up tree if no more siblings */
				current = next;
				x -= tree[current].r.x;
				y -= tree[current].r.y;
				next = tree[current].ob_next;
			}
			current = next;
		}
	} while(current != -1);		/* If 'current' is -1 then we have finished */

	return 0;	/* Bummer - didn't find the object, so return error */
}

/*
 * Find which object is at a given location
 */
short find_object(OBJECT *tree, short object, short depth, short obx, short oby)
{
	short next;
	short current = 0, rel_depth = 1;
	short x = 0, y = 0, start_checking = FALSE;
	short pos_object = -1;
	
	do
	{
		if (current == object)	/* We can start considering objects at this point */
		{
			start_checking = TRUE;
			rel_depth = 0;
		}
		
		if (start_checking)
		{
			if (   (tree[current].ob_flags&HIDETREE) == 0		/* HR */
				&& (tree[current].r.x + x <= obx)
				&& ((tree[current].r.y + y <= oby)
				&& ((tree[current].r.x + x + tree[current].r.w >= obx)
				&& (tree[current].r.y + y + tree[current].r.h >= oby))))
			{
				pos_object = current;	/* This is only a possible object, as it may have children on top of it. */
			}
		}

		if (   (   (!start_checking)
		        || (rel_depth < depth)
		       )
		    && (tree[current].ob_head != -1)
		    && (tree[current].ob_flags&HIDETREE) == 0			/* HR */
		   )
		{		/* Any children? */
			x += tree[current].r.x;
			y += tree[current].r.y;
			rel_depth++;
			current = tree[current].ob_head;
		} else
		{
			next = tree[current].ob_next;							/* Try for a sibling */

			while((next != -1) && (tree[next].ob_tail == current))	/* Trace back up tree if no more siblings */
			{
				current = next;
				x -= tree[current].r.x;
				y -= tree[current].r.y;
				next = tree[current].ob_next;
				rel_depth--;
			}
			current = next;
		}
		
	} while((current != -1) && (rel_depth > 0));

	return pos_object;
}

#if 1
/*
 *	Perform a few fixes on a menu tree prior to installing it
 *	(ensure title spacing & items fit into their menus)
 */

/* HR: N.B. Using standard draw_object_tree for the menu bar has many advantages :-)
            There is however 1 disadvantage: TOS doesnt do this, so we may expect some
            incompatabilities.
*/

/* HR: Removed all spacing corrections. */
void fix_menu(OBJECT *root, bool desk)		/* Called by menu_bar() only */
{
	short titles, menus, tbar, s_ob, t_ob;
	unsigned long cl;

	DIAG((D.menu,-1,"fixing menu @%lx\n", root));

	cl =  menu_colors() | (1 << 7);		/* | transparent text */

	tbar = root[0].ob_head;
	titles = root[tbar].ob_head;
	menus = root[0].ob_tail;

/* HR: */
	root->r.w = root[tbar].r.w = root[menus].r.w = display.r.w;
	set_obspec(root,tbar,cl);

	root[menus].ob_flags |= HIDETREE;	/* Hide the actual menus (The big ibox) */

	t_ob = root[titles].ob_head;
	s_ob = root[menus].ob_head;

	while(t_ob != titles)
	{
		root[s_ob].ob_flags |= HIDETREE;	/* Hide the actual menu */
		set_obspec(root,s_ob,0xff0000L | cl);
		t_ob = root[t_ob].ob_next;
		s_ob = root[s_ob].ob_next;
	}


/*  fix desk menu */
	if (desk)
	{
		s_ob = root[menus].ob_head;
		t_ob = root[s_ob ].ob_head;
		root[s_ob].r.h = root[t_ob+3].r.y - root[s_ob].r.y;
		t_ob+=2;
		root[t_ob++].ob_spec = "  Clients \3";
		while (t_ob != s_ob)
		{
			root[t_ob].ob_flags |= HIDETREE|DISABLED;
			t_ob = root[t_ob].ob_next;
		}
	}
	DIAG((D.menu,-1,"done fix_menu()\n"));
}

#else
#include "obsolete/fix_menu.h"
#endif

/*
 *	Object Tree Handling Interface
 */

AES_function XA_objc_draw		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	RECT *r = (RECT *)&pb->intin[2];

	CONTROL(6,1,1)

#if GENERATE_DIAGS
	global_pid_for_draw = client->pid;
#endif
	v_hide_c(V_handle);
	set_clip(r);
	pb->intout[0] = draw_object_tree(pb->addrin[0], pb->intin[0], pb->intin[1] + 1);
	clear_clip();
	v_show_c(V_handle, 1);
	
	return XAC_DONE;
}

AES_function XA_objc_offset	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(1,3,1)

	pb->intout[0] = object_abs_coords(pb->addrin[0], pb->intin[0], pb->intout + 1, pb->intout + 2);

	return XAC_DONE;
}

AES_function XA_objc_find		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(4,1,1)

	pb->intout[0] = find_object(pb->addrin[0], pb->intin[0], pb->intin[1], pb->intin[2], pb->intin[3]);

	return XAC_DONE;
}

/* HR: the whole tree must be redrawn, otherwise transparents objects go wrong!
 *     depth must be some default maximum
 * Care must be taken that outside borders or shadows are included in the draw.
 * Only the objects area is redrawn, so it must be intersected with clipping rectangle.
 * New function object_area(RECT *c, tree, item)
 *
 *	Now we can use this for the standard menu's and titles!!!
 */
#if 1

/* Preliminary: (See if AHCX's window menu's are drawn correctly (no borders there :-) */

short thickness(OBJECT *ob)
{
	short t = 0;
	switch(ob->ob_type&0xff)
	{
	case G_BOX:
	case G_IBOX:
	case G_BOXCHAR:
		t = (((unsigned long)ob_spec(ob)) & 0xff0000L) >> 16;
	}
	if (t & 128)
		t = -(1 + (t ^ 0xff));
	return t;
}

int max(short a, short b)
{
	return a > b ? a : b;
}

int min(short a, short b)
{
	return a < b ? a : b;
}

void object_area(RECT *c,OBJECT *ob, short i)
{
	OBJECT *b = ob + i;
	short border_thick;
	short dx=0,dy=0,dw=0,dh=0,db;
	object_abs_coords(ob, i, &c->x, &c->y);
	c->w = b->r.w;
	c->h = b->r.h;

    db = border_thick = thickness(ob+i);   /* type dependent */
	
	/*	Are we shadowing this object? (Borderless objects aren't shadowed!) */
	if (border_thick < 0) 	/* outside border */
	{
		db = -border_thick;
		dx = border_thick;
		dy = border_thick;
		dw = 2*db;
		dh = 2*db;
	}
	if (b->ob_state & SHADOWED)
	{
		dw += 2*db;
		dh += 2*db;
	}

	if (b->ob_state & OUTLINED)
	{
		dx = min(dx, -3);
		dy = min(dy, -3);
		dw = max(dw,  6);
		dh = max(dh,  6);
	}

	c->x += dx;
	c->y += dy;
	c->w += dw;
	c->h += dh;
}

bool check_tree(XA_CLIENT *client, OBJECT *tree, short item);

void change_object(OBJECT *root, short i, short start, RECT *r, short state, short draw)
{
/*	DIAGS(("change_object %lx, %d\n", root, i)); */

	root[i].ob_state = state;

	if (draw)
	{
		RECT c; short t = root[i].ob_type & 0xff;

/* A quick hack to catch *most* of the problems with transparent objects */
		if (i == start)
			if (t == G_STRING || t == G_SHORTCUT || t == G_TITLE)
			{
				short q = get_parent(root,i);
				if (q >= 0 && t == G_TITLE)
					q = get_parent(root,q);
				if (q >= 0)
					start = q;
			}
/* */
		v_hide_c(V_handle);
		object_area(&c,root,i);
		if (   !r
			|| (r && rc_intersect(r,&c))
			)
		{
			set_clip(&c);
			draw_object_tree(root, start, 100);
			clear_clip();
		}
		v_show_c(V_handle, 1);
	}
}

AES_function XA_objc_change	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(8,1,1)

	change_object((OBJECT *)pb->addrin[0],
							pb->intin[0],
/* The GEM version doesnt distinguish between a starting item for draw and the item 
   that is to be changed. To my experience using the same for draw and start is 
   what current AES's do. (Otherwise see what Luna.app does when selecting it's own menu items).
*/
							pb->intin[0],
							(RECT *)&pb->intin[2],
							pb->intin[6],
							pb->intin[7]  );	
	pb->intout[0] = 1;
	
	return XAC_DONE;
}
#else
AES_function XA_objc_change	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	OBJECT *root = (OBJECT *)pb->addrin[0];

	CONTROL(8,1,1)

	root[pb->intin[0]].ob_state = pb->intin[6];
	
	if (pb->intin[7])
	{
		v_hide_c(V_handle);
		set_clip((RECT *)&pb->intin[2]);
		draw_object_tree(root, pb->intin[0], pb->intin[1] + 1);
		clear_clip();
		v_show_c(V_handle, 1);
	}
	
	pb->intout[0] = 1;
	
	return XAC_DONE;
}
#endif

AES_function XA_objc_add	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	OBJECT *root = (OBJECT *)pb->addrin[0];
	short parent, new_child, old_child;
	
	CONTROL(2,1,1)

	parent = pb->intin[0];
	old_child = root[parent].ob_head;
	new_child = pb->intin[1];
	
	if (old_child == -1)	/* Add new object first among siblings */
	{
		root[new_child].ob_next = parent;
		root[parent].ob_tail = new_child;
	} else
		root[new_child].ob_next = old_child;
	
	root[parent].ob_head = new_child;

	pb->intout[0] = 1;
	
	return XAC_DONE;
}

static short remove_object(OBJECT *root, short object)
{
	short parent, current, last;

	current = object;
	do
	{						/* Find parent */
		last = current;
		current = root[current].ob_next;
	} while (root[current].ob_tail != last);
	parent = current;
	
	if (root[parent].ob_head == object)	/* First child */
	{
		if (root[object].ob_next == parent)		/* No siblings */
			root[parent].ob_head = root[parent].ob_tail = -1;
		else						/* Siblings */
			root[parent].ob_head = root[object].ob_next;
	} else								/* Not first child */
	{
		current = root[parent].ob_head;
		do
		{						/* Find adjacent sibling */
			last = current;
			current = root[current].ob_next;
		} while (current != object);
		root[last].ob_next = root[object].ob_next;
		if (root[object].ob_next == parent)
			root[parent].ob_tail = last;	/* Last child removed */
	}
	
	return parent;
}

AES_function XA_objc_delete	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(1,1,1)

	remove_object((OBJECT *)pb->addrin[0], pb->intin[0]);

	pb->intout[0] = 1;

	return XAC_DONE;
}

AES_function XA_objc_order	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	OBJECT *root = (OBJECT *)pb->addrin[0];
	short parent, object, current;
	unsigned short pos;		/* -1 (top) is a large positive number */
	
	CONTROL(2,1,1)

	object = pb->intin[0];
	parent = remove_object(root, object);

	pos = pb->intin[1];	
	current = root[parent].ob_head;
	if (current == -1)		/* No siblings */
	{
		root[parent].ob_head = root[parent].ob_tail = object;
		root[object].ob_next = parent;
	} else if (!pos)	/* First among siblings */
	{
		root[object].ob_next = current;
		root[parent].ob_head = object;
	} else				/* Search for position */
	{
		for(pos--; pos && root[current].ob_next != parent; pos--)
			current = root[current].ob_next;
		if (root[current].ob_next == parent)
			root[parent].ob_tail = object;
		root[object].ob_next = root[current].ob_next;
		root[current].ob_next = object;
	}

	pb->intout[0] = 1;

	return XAC_DONE;
}



/* Johan's versions of these didn't work on my system, so I've redefined them 
   - This is faster anyway */

/* HR keep this at the end of the file */
#define CLIP(tree,obj,r)	clear_clip()
#define set_clip(x)		/* We don't want this now! */

const unsigned char character_type[] =
{
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	CGs, 0, 0, 0, 0, 0, 0, 0, 0, 0, CGw, 0, 0, 0, CGdt, 0,
	CGd, CGd, CGd, CGd, CGd, CGd, CGd, CGd, CGd, CGd, CGp, 0, 0, 0, 0, CGw,
	0, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa,
	CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, CGu|CGa, 0, CGp, 0, 0, CGxp,
	0, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa,
	CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, CGa, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#if WRONG_LEN
short ed_char(TEDINFO *ed_txt, short keycode)
{
	char *txt = ed_txt->te_ptext;
	short cursor_pos = ed_txt->te_tmplen,
		  o, x, key, tmask, n, chg, update = 0;

	switch(keycode)
	{	
	case 0x011b:		/* ESCAPE clears the field */
		txt[0] = '\0';
		ed_txt->te_tmplen = 0;
		update = 1;
		break;

	case 0x537f:		/* DEL deletes character under cursor */
		if (txt[cursor_pos])
		{
			for (x = cursor_pos; x < ed_txt->te_txtlen - 1; x++)
				txt[x] = txt[x + 1];
			
			update = 1;
		}
		break;
			
	case 0x0e08:		/* BACKSPACE deletes character behind cursor (if any) */
		if (cursor_pos)
		{
			for (x = cursor_pos; x < ed_txt->te_txtlen; x++)
				txt[x - 1] = txt[x];
					
			ed_txt->te_tmplen--;

			update = 1;
		}
		break;
				
	case 0x4d00:	/* RIGHT ARROW moves cursor right */
		if ((txt[cursor_pos]) && (cursor_pos < ed_txt->te_txtlen - 1))
		{
			ed_txt->te_tmplen++;
			update = 1;
		}
		break;

	case 0x4d36:	/* SHIFT+RIGHT ARROW move cursor to far right of current text */
		for (x = 0; txt[x]; x++)
			;

		if (x != cursor_pos)
		{
			ed_txt->te_tmplen = x;
			update = 1;
		}
		break;
			
	case 0x4b00:	/* LEFT ARROW moves cursor left */
		if (cursor_pos)
		{
			ed_txt->te_tmplen--;
			update = 1;
		}
		break;
			
	case 0x4b34:	/* SHIFT+LEFT ARROW move cursor to start of field */
	case 0x4700:	/* CLR/HOME also moves to far left */
		if (cursor_pos)
		{
			ed_txt->te_tmplen = 0;
			update = 1;
		}
		break;

	default:		/* Just a plain key - insert character */
		chg = 0;		/* Ugly hack! */
		if (cursor_pos == ed_txt->te_txtlen - 1)
		{
			cursor_pos--;
			ed_txt->te_tmplen--;
			chg = 1;
		}
				
		key = keycode & 0xff;
		tmask=character_type[key];

		n = strlen(ed_txt->te_pvalid) - 1;
		if (cursor_pos < n)
			n = cursor_pos;

		switch(ed_txt->te_pvalid[n])
		{
		case '9':
			tmask &= CGd;
			break;
		case 'a':
			tmask &= CGa|CGs;
			break;
		case 'n':
			tmask &= CGa|CGd|CGs;
			break;
		case 'p':
			tmask &= CGa|CGd|CGp|CGxp;
			/*key = toupper((char)key);*/
			break;
		case 'A':
			tmask &= CGa|CGs;
			key = toupper((char)key);
			break;
		case 'N':
			tmask &= CGa|CGd|CGs;
			key = toupper((char)key);
			break;
		case 'F':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'f':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'P':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'X':
			tmask = 1;
			break;
		case 'x':
			tmask = 1;
			key = toupper((char)key);
			break;
		default:
			tmask = 0;
			break;			
		}
		
		if (!tmask)
		{
			for (n = x = 0; ed_txt->te_ptmplt[n]; n++)
			{
		   		if (ed_txt->te_ptmplt[n] == '_')
					x++;
				else if ((ed_txt->te_ptmplt[n] == key)
					&& (x >= cursor_pos))
					break;
			}
			if (key && (ed_txt->te_ptmplt[n] == key))
			{
				for (n = cursor_pos; n < x; n++)
					txt[n] = ' ';
				txt[x] = '\0';
				ed_txt->te_tmplen = x;
			} else
			{
				ed_txt->te_tmplen += chg;		/* Ugly hack! */
				return XAC_DONE;
			}
		} else
		{
			txt[ed_txt->te_txtlen - 2] = '\0';	/* Needed! */
			for (x = ed_txt->te_txtlen - 1; x > cursor_pos; x--)
				txt[x] = txt[x - 1];

			txt[cursor_pos] = (char)key;

			ed_txt->te_tmplen++;
		}

		update = 1;
		break;
	}

	return update;
}
#else
short ed_char(XA_WIDGET_TREE *wt, TEDINFO *ed_txt, short keycode)
{
	char *txt = ed_txt->te_ptext;
	short cursor_pos = wt->edit_pos,
		  o, x, key, tmask, n, chg, update = 0;

	switch(keycode)
	{	
	case 0x011b:		/* ESCAPE clears the field */
		txt[0] = '\0';
		update = 1;
		break;

	case 0x537f:		/* DEL deletes character under cursor */
		if (txt[cursor_pos])
		{
			for(x = cursor_pos; x < ed_txt->te_txtlen - 1; x++)
				txt[x] = txt[x + 1];
			
			update = 1;
		}
		break;
			
	case 0x0e08:		/* BACKSPACE deletes character left of cursor (if any) */
		if (cursor_pos)
		{
			for(x = cursor_pos; x < ed_txt->te_txtlen; x++)
				txt[x - 1] = txt[x];

			update = 1;
		}
		break;
				
	case 0x4d00:	/* RIGHT ARROW moves cursor right */
		if ((txt[cursor_pos]) && (cursor_pos < ed_txt->te_txtlen - 1))
		{
			cursor_pos++;
			update = 1;
		}
		break;

	case 0x4d36:	/* SHIFT+RIGHT ARROW move cursor to far right of current text */
		for(x = 0; txt[x]; x++)
			;

		if (x != cursor_pos)
		{
			cursor_pos = x;
			update = 1;
		}
		break;
			
	case 0x4b00:	/* LEFT ARROW moves cursor left */
		if (cursor_pos)
		{
			cursor_pos--;
			update = 1;
		}
		break;
			
	case 0x4b34:	/* SHIFT+LEFT ARROW move cursor to start of field */
	case 0x4700:	/* CLR/HOME also moves to far left */
		if (cursor_pos)
		{
			cursor_pos = 0;
			update = 1;
		}
		break;

	default:		/* Just a plain key - insert character */
		chg = 0;		/* Ugly hack! */
		if (cursor_pos == ed_txt->te_txtlen - 1)
		{
			cursor_pos--;
			chg = 1;
		}
				
		key = keycode & 0xff;
		tmask=character_type[key];

		n = strlen(ed_txt->te_pvalid) - 1;
		if (cursor_pos < n)
			n = cursor_pos;

		switch(ed_txt->te_pvalid[n])
		{
		case '9':
			tmask &= CGd;
			break;
		case 'a':
			tmask &= CGa|CGs;
			break;
		case 'n':
			tmask &= CGa|CGd|CGs;
			break;
		case 'p':
			tmask &= CGa|CGd|CGp|CGxp;
			/*key = toupper((char)key);*/
			break;
		case 'A':
			tmask &= CGa|CGs;
			key = toupper((char)key);
			break;
		case 'N':
			tmask &= CGa|CGd|CGs;
			key = toupper((char)key);
			break;
		case 'F':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'f':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'P':
			tmask &= CGa|CGd|CGp|CGxp|CGw;
			/*key = toupper((char)key);*/
			break;
		case 'X':
			tmask = 1;
			break;
		case 'x':
			tmask = 1;
			key = toupper((char)key);
			break;
		default:
			tmask = 0;
			break;			
		}
		
		if (!tmask)
		{
			for(n = x = 0; ed_txt->te_ptmplt[n]; n++)
			{
		   		if (ed_txt->te_ptmplt[n] == '_')
					x++;
				else if ((ed_txt->te_ptmplt[n] == key)
					&& (x >= cursor_pos))
					break;
			}
			if (key && (ed_txt->te_ptmplt[n] == key))
			{
				for(n = cursor_pos; n < x; n++)
					txt[n] = ' ';
				txt[x] = '\0';
				cursor_pos = x;
			} else
			{
				cursor_pos += chg;		/* Ugly hack! */
				wt->edit_pos = cursor_pos;
				return XAC_DONE;
			}
		} else
		{
			txt[ed_txt->te_txtlen - 2] = '\0';	/* Needed! */
			for(x = ed_txt->te_txtlen - 1; x > cursor_pos; x--)
				txt[x] = txt[x - 1];

			txt[cursor_pos] = (char)key;

			cursor_pos++;
		}

		update = 1;
		break;
	}
	wt->edit_pos = cursor_pos;
	return update;
}
#endif

/* HR: Only the '\0' in the te_ptext field is determining the corsor position, NOTHING ELSE!!!!
		te_tmplen is a constant and strictly owned by the APP.
*/

AES_function XA_objc_edit		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	OBJECT *form = pb->addrin[0];
	short ed_obj = pb->intin[0];
	short keycode = pb->intin[1];
	TEDINFO *ed_txt;
	short  o, update = 0;
	RECT r;

	CONTROL(4,2,1)

	ed_txt = (TEDINFO *)ob_spec(&form[ed_obj]);

	pb->intout[0] = 1;

	switch(pb->intin[3])
	{
	case 0:
		break;
			
	case 1:			/* ED_INIT - set current edit field */
#if 0				/* For some reason, Johan's version here crashes CAB. Not sure why, so I've put my old version back for now */
					/* HR: I think I know why!!!! te_tmplen is supposed to be a constant!!! */
		o = 0;
		do
			form[o].ob_state &= ~IS_EDIT;
		while( ! (form[o++].ob_flags & LASTOB));	/* HR: Check LASTOB before incrementing Could this be the cause ??? of the crash? */
			
		form[ed_obj].ob_state |= IS_EDIT;
	#if WRONG_LEN
		ed_txt->te_tmplen = strlen(((TEDINFO *)ob_spec(&form[ed_obj]))->te_ptext);
	#endif
		update = 1;
#else
		o = 0;
		do
			form[o].ob_state &= ~IS_EDIT;
		while( ! (form[o++].ob_flags & LASTOB));	/* HR: Check LASTOB before incrementing */
			
		form[ed_obj].ob_state |= IS_EDIT;
	#if WRONG_LEN
		((TEDINFO *)ob_spec(&form[ed_obj]))->te_tmplen = 0;
	#endif
		object_abs_coords(form, ed_obj, &r.x, &r.y);
		r.w = (form + ed_obj)->r.w;
		r.h = (form + ed_obj)->r.h;
		set_clip(&r);
				
		v_hide_c(V_handle);
		draw_object_tree(form, ed_obj, 2);
		v_show_c(V_handle, 1);

		pb->intout[1] = 0;

#endif

		break;

	case 2:			/* ED_CHAR - process a character */
		update = ed_char(ed_txt,keycode);
#if WRONG_LEN
		pb->intout[1] = ed_txt->te_tmplen;
#else
		pb->intout[1] = strlen(ed_txt->te_ptext);
#endif
		break;
			
	case 3:		/* ED_END - turn off the cursor */
		form[ed_obj].ob_state &= ~IS_EDIT;
		update = 1;
		break;
	}
	if (update)			/* Moved from a number of places above. */
	{
		CLIP(form, ed_obj, r);
		v_hide_c(V_handle);
		draw_object_tree(form, ed_obj, MAX_DEPTH);
		v_show_c(V_handle, 1);
#if WRONG_LEN
		pb->intout[1] = ed_txt->te_tmplen;
#else
		pb->intout[1] = strlen(ed_txt->te_ptext);
#endif
	}
	
	return XAC_DONE;
}

/* HR: objc_sysvar
   I think some programs need this to be able to do extended
    objects in 3D using progdefs.
    And shure they have the right to call this when they can do
    3D objects at all.    (appl_getinfo).
    .....
    I was right: Hip hip hurray!!! Luna.app buttons now OK.
    Now many more modern programs will behave correct.
*/
 
AES_function XA_objc_sysvar	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
DIAG((D.objc,-1,"objc_sysvar %s: %d for %d\n", pb->intin[0] == SV_INQUIRE ? "inq" : "set", pb->intin[1], client->pid));

	CONTROL(4,3,0)

	if (pb->intin[0] == SV_INQUIRE)		/* SV_SET later: or not??? */
	{
		switch(pb->intin[1])
		{
		case LK3DIND:
			pb->intout[1] = 0;		/* text move (selecting) */
			pb->intout[2] = 0;		/* color change (selecting) */
			break;
		case LK3DACT:
			pb->intout[1] = 1;		/* text move (selecting) */
			pb->intout[2] = 1;		/* color change (selecting) */
			break;
		case INDBUTCOL:
			pb->intout[1] = display.dial_colours.bg_col;
			break;
		case ACTBUTCOL:
			pb->intout[1] = display.dial_colours.bg_col;
			break;
		case BACKGRCOL:
			pb->intout[1] = display.dial_colours.bg_col;
			break;
		case AD3DVAL:
			pb->intout[1] = MONO ? 2 : 1;  /* XA_3D_pushbutton enlargement values :-) */
			pb->intout[2] = MONO ? 2 : 1;
			break;
		default:
			pb->intout[0] = 0;
			return XAC_DONE;
		}
	}

	pb->intout[0] = 1;

	return XAC_DONE;
}
