/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <osbind.h>
#include <memory.h>
#include <string.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "xa_rsrc.h"
#include "objects.h"		/* HR */

/*
 *	RESOURCE FILE HANDLER
 *
 *	Simulate the standard GEM AES resource access functions.
 *	
 *	I've added these to Steve's routines to act as a bit of an interface
 *	to his routines. Each application has its own std_resource as part
 *	of the XA_CLIENT structure. As Steve's resource loader is happy to
 *	allow multiple resource files, I should add some extra calls to support it.
 */


/*
 * Fixup OBJECT coordinates; the lower byte contains a (character-based)
 * coordinate, the upper byte an additional (pixel) offset.
 */
#define fixup(val,scale) (((val) & 0xff) * (scale) + (((val) >> 8) & 0xff))

void obfix(OBJECT *tree, short object)
{
	OBJECT *o = &tree[object];

	o->r.x = fixup(o->r.x, display.c_max_w) ;
	o->r.y = fixup(o->r.y, display.c_max_h) ;
	/*
	 * Special case handling: any OBJECT 80 characters wide is supposed
	 * to be a menu bar, which always covers the entire screen width...
	 */
	o->r.w = (o->r.w == 80) ?
			display.r.w : fixup(o->r.w, display.c_max_w) ;
	o->r.h = fixup(o->r.h, display.c_max_h) ;
}

/*
 * Code in this module is based on the resource loader from
 * Steve Sowerby's AGiLE library. Thanks to Steve for allowing
 * me to use his code.
 */

short resWidth, resHeight;

/*
 * FixColourIconData : Convert a colour icon from device independent to device specific
 */
static void FixColourIconData(XA_CLIENT *client, CICONBLK *icon, long base)	/* base is usefull for debug info */
{
	CICON *c;
	MFDB src, dst;
	unsigned long len, icon_len, new_len;
	short *new_data;
	short *tmp = 0;

	len = calc_back(&icon->monoblk.ic, 1);		/* HR: use functions !! */

DIAG((D.rsrc,-1,"color icon: '%s' %d*%d %ld tx.w=%d\n",
		icon->monoblk.ib_ptext,
		icon->monoblk.ic.w, icon->monoblk.ic.h, len, icon->monoblk.tx.w));

	c = icon->mainlist;
	while (c)
	{
		icon_len  = c->num_planes * len;
		new_len  = display.planes * len;

DIAG((D.rsrc,-1,"icon_len %ld, new_len %ld\n", icon_len, new_len));
		if (c->num_planes < display.planes)	/* Create padded data if necessary */
		{
			if (!c->sel_data)
			{
			/*	new_data = Mxalloc(new_len,  MX_GLOBAL | MX_PREFTTRAM);		/* HR global*/	/* Why ST-RAM? */
			*/  new_data = XA_alloc(client, new_len, 0, client->rsct);		/* HR: and put in a list. */
				memcpy(        new_data,  c->col_data, icon_len);
				memset((char *)new_data + icon_len, 0, new_len - icon_len);
			} else
			{
			/*	new_data = Mxalloc(new_len * 2,  MX_GLOBAL | MX_PREFTTRAM);	/* HR global*/	/* Why ST-RAM?  */
			*/	new_data = XA_alloc(client, new_len * 2, 0, client->rsct);	/* HR: and put in a list. */
				memcpy(        new_data,            c->col_data, icon_len);
				memset((char *)new_data + icon_len, 0, new_len - icon_len);
				memcpy((char *)new_data + new_len,  c->sel_data, icon_len);
				memset((char *)new_data + new_len + icon_len, 0, new_len - icon_len);
				c->sel_data = (short *)((char *)new_data + new_len);
			}
			c->col_data = new_data;
		}

		src.fd_w = icon->monoblk.ic.w;			/* Transform MFDB's */
		src.fd_h = icon->monoblk.ic.h;
		src.fd_wdwidth = (src.fd_w + 15) / 16;		/* HR: round up */
		src.fd_stand = 1;
		src.fd_r1 = src.fd_r2 = src.fd_r3 = 0;
		src.fd_nplanes = display.planes /* <= 8 ? display.planes : c->num_planes */;		/* HR */
		dst = src;

		dst.fd_addr = c->col_data;
		dst.fd_stand = 0;

		if (icon_len > MAX_IN_PLACE_TRNFM)
		{								/* HR: tmp is a local zeroised pointer */
			tmp = calloc(1,new_len);		/* HR: bad bug!!  was icon_len + 10 ?? (too small!!) */
			memcpy(tmp, c->col_data, new_len);
			src.fd_addr = tmp;
		} else
			src.fd_addr = c->col_data;

		vr_trnfm(V_handle, &src, &dst);

		if (c->sel_data)
		{
			dst.fd_addr = c->sel_data;
			if (icon_len > MAX_IN_PLACE_TRNFM)
				memcpy(tmp, c->sel_data, new_len);	/* HR: was icon_len */			
			else
				src.fd_addr = c->sel_data;
			vr_trnfm(V_handle, &src, &dst);
		}
/*		if (display.planes > 8)		/* transform in place to higher colors system */
		{
			src.fd_nplanes = display.planes;
			dst.fd_nplanes = display.planes;
			src.fd_addr = c->col_data;
			dst.fd_addr = c->col_data;
			src.fd_stand = 0;
			dst.fd_stand = 0;
			vr_trnfm(V_handle, &src, &dst);
			if (c->sel_data)
			{
				src.fd_addr = c->sel_data;
				dst.fd_addr = c->sel_data;
				vr_trnfm(V_handle, &src, &dst);
			}
		}
*/		c = c->next_res;
	}
	
	if (tmp)
		free(tmp);
}


/*
 * LoadResources : Load a GEM resource file
 * fname = name of file to load
 * Return = base pointer of resources or NULL on failure
 */
void *LoadResources(XA_CLIENT *client, char *fname, short designWidth, short designHeight)
{
	long err;
	short handle, pid = client->pid;
	RSHDR hdr;
	OBJECT *obj ,*tree, **trees;
	CICONBLK **cibh;
	unsigned long osize, size;
	char *base;
	short i, j, type, numCibs = 0;

	resWidth = display.c_max_w;
	resHeight = display.c_max_h;
	
	DIAG((D.rsrc,pid,"LoadResources(%s)\n", fname));
	
	err = Fopen(fname, O_RDONLY);
	if (err < 0L)
	{
		DIAG((D.rsrc,pid,"LoadResources(): file not found\n"));
		return NULL;
	}
	handle = (short)err;
	Fread(handle, sizeof(RSHDR), &hdr);
	size = (unsigned long)hdr.rsh_rssize;
	osize = (size + 1UL) & 0xfffffffeUL;
	if (hdr.rsh_vrsn & 4)	/* Extended format, get new 32-bit length */
	{
		Fseek(osize,handle, SEEK_SET);
		Fread(handle, 4L, &size);
		DIAG((D.rsrc,pid,"extended format osize=%ld: read extended size=%ld\n", osize, size));
	}

	Fseek(0L, handle, SEEK_SET);

/*	base = Mxalloc(size, MX_GLOBAL | MX_PREFTTRAM);		/* HR global*/
*/	base = XA_alloc(client, size, 1, client->rsct + 1);	/* HR: and put in a list. */
	if (!base)
	{
		DIAG((D.rsrc,pid,"Can't allocate space for resource file\n"));
		Fclose(handle);
		return NULL;
	}

	client->rsct++;
	Fread(handle, size, base);			/* Reread everything */
	Fclose(handle);

	{	/* fixup all free string pointers */
		char **fs = (char **)(base + hdr.rsh_frstr);
		for (i = 0; i < hdr.rsh_nstring; i++, fs++)
		{
			char *d = *fs;
			*fs += (long)base;
			DIAG((D.s,pid,"fs[%d]>%ld='%s'\n",i,d,*fs));
		}
		
		DIAG((D.rsrc,pid,"fixed up %d free_string pointers\n",hdr.rsh_nstring));
	}
	{	/* HR: fixup all free image pointers */
		char **fs = (char **)(base + hdr.rsh_frimg);
		for (i = 0; i < hdr.rsh_nimages; i++, fs++)
		{
			char *d = *fs;
			*fs += (long)base;
			DIAG((D.rsrc,pid,"imgs[%d]>%ld=%lx\n",i,d,*fs));
		}
		
		DIAG((D.rsrc,pid,"fixed up %d free_image pointers\n",hdr.rsh_nimages));
	}
	{	/* fixup all tedinfo field pointers */
		TEDINFO *ti = (TEDINFO *)(base + hdr.rsh_tedinfo);
		for (i = 0; i < hdr.rsh_nted; i++, ti++)
		{
			(long)ti->te_ptext  += (long)base;
			(long)ti->te_ptmplt += (long)base;
			(long)ti->te_pvalid += (long)base;
		}
	
		DIAG((D.rsrc,pid,"fixed up %d tedinfo's\n",hdr.rsh_nted));
	}
	{	/* fixup all iconblk field pointers */
		ICONBLK *ib = (ICONBLK *)(base + hdr.rsh_iconblk);
		for (i = 0; i < hdr.rsh_nib; i++, ib++)
		{
			(long)ib->ib_pmask += (long)base;
			(long)ib->ib_pdata += (long)base;
			(long)ib->ib_ptext += (long)base;
		}
		
		DIAG((D.rsrc,pid,"fixed up %d iconblk's\n", hdr.rsh_nib));
	}
	{	/* fixup all bitblk data pointers */
		BITBLK *bb = (BITBLK *)(base + hdr.rsh_bitblk);
		for (i = 0; i < hdr.rsh_nbb; i++, bb++)
		{
			(long)bb->bi_pdata += (long)base;
		}

		DIAG((D.rsrc,pid,"fixed up %d bitblk's\n", hdr.rsh_nbb));
	}

	if (hdr.rsh_vrsn & 4)	/* It's an enhanced RSC file */
	{
		unsigned long *earray, *addr;

		DIAG((D.rsrc,pid,"Enhanced resource file\n"));

		earray = (unsigned long *)(osize + (long)base);		/* this the row of pointers to extensions */
															/*          terminated by a 0L */
		cibh = (CICONBLK **)(earray[1] + (long)base);
		if ((long)cibh > 0L)	/* Get colour icons */
		{
			CICONBLK *cib, **cp = cibh;

			while (*cp++ != (CICONBLK *) -1L)
				numCibs++;

			cib = (CICONBLK *)cp;

			for (i = 0; i < numCibs; i++)	/* Fix up all the CICONBLK's */
			{
				CICON *cicn;
				ICONBLK *ib = &cib->monoblk;
				short *pdata, numRez;

				cibh[i] = cib;
		/*		size = ((ib->ic.w+15) / 16) * ib->ic.h * 2;		/* HR: round up (Very bad bug!!!) */
		*/		size = calc_back(&ib->ic, 1);					/*     use functions! once good, always good. */
				addr = (unsigned long*)((long)cib + sizeof(ICONBLK));
				numRez = addr[0];
				pdata = (short *)&addr[1];
				ib->ib_pdata = pdata;						/* mono data & mask */
				(long)pdata += size;
				ib->ib_pmask = pdata;
				(long)pdata += size;
/* HR: the texts are placed the same way as for all other objects
		when longer than 12 bytes.		*/
				if (ib->ib_ptext)
				{
					short l = ib->tx.w/6;
					ib->ib_ptext += (long)base;		/* HR: fix some resources */
					if (strlen(ib->ib_ptext) > l)
						*(ib->ib_ptext + l) = 0;
				} else
/*	HR: The following word is no of planes which cannot be larger than 32, so there is 1 zero byte there */
					ib->ib_ptext = (char *)pdata; 

				(long)pdata += 12;						/* (unused) place for name */

				cicn = (CICON *)pdata;
/* HR: There can be color icons with NO color icons, only the mono icon block. */
				cib->mainlist = NULL;
				for (j = 0; j < numRez; j++)	/* Get CICON's at different rez's */
				{
					short planes = cicn->num_planes;
					long psize = size * planes;
					(long)pdata += sizeof(CICON);
					cicn->col_data = pdata;
					(long)pdata += psize;
					cicn->col_mask = pdata;
					(long)pdata +=  size;
					if (cicn->sel_data)				/* It's got a selected form */
					{
						cicn->sel_data = pdata;
						(long)pdata += psize;
						cicn->sel_mask = pdata;
						(long)pdata +=  size;
					} else	/* No selected version */
						cicn->sel_data = cicn->sel_mask = NULL;

/* HR */			if (cib->mainlist == NULL)
						cib->mainlist = cicn;

					if ((long)cicn->next_res == 1)
						cicn->next_res = (CICON *)pdata;
					else
						cicn->next_res = NULL;

					cicn = (CICON *)pdata;
				}
				cib = (CICONBLK *)cicn;
			}
			DIAG((D.rsrc,pid,"fixed up %d color icons\n", numCibs));
		}
	}

/* HR: As you see, the objects are NOT in a pointer array!!! */
	obj = (OBJECT *)(base + hdr.rsh_object);
	tree = obj;
	for (i = 0; i < hdr.rsh_nobs; i++, obj++)	/* fixup all objects' ob_spec pointers */
	{
		type = obj->ob_type & 255;
DIAG((D.s,pid,"obj[%d]>%ld=%lx, %s;\t%d,%d,%d\n",
						i,
						(long)obj-(long)base,
						obj,
						object_type(tree,i),
						obj->ob_next,obj->ob_head,obj->ob_tail));
		switch (type) 	/* What kind of object is it? */
		{
		case G_TEXT:
		case G_BOXTEXT:
		case G_IMAGE:
		case G_BUTTON:
		case G_STRING:
		case G_SHORTCUT:	/* HR */
		case G_FTEXT:
		case G_FBOXTEXT:
		case G_ICON:
		case G_TITLE:
			obj->ob_spec = (void *)((unsigned long)obj->ob_spec + base);
			break;
		case G_CICON:
			FixColourIconData(client, cibh[(long)obj->ob_spec], (long)base);
			obj->ob_spec = cibh[(long)obj->ob_spec];
			break;
		case G_PROGDEF:
			obj->ob_spec = NULL;
			break;
		case G_IBOX:
		case G_BOX:
		case G_BOXCHAR:
			break;
		default:
			DIAG((D.rsrc,pid,"Unknown object type %d\n", type));
			break;
		}
	}

	DIAG((D.rsrc,pid,"fixed up %d objects ob_spec\n", hdr.rsh_nobs));

/* HR!!!!! the pointer array wasnt fixed up!!!!! which made it unusable via global[5] */

	(unsigned long)trees = (unsigned long)(base + hdr.rsh_trindex);		/* HR: OBJECT **trees */
	for (i = 0; i < hdr.rsh_ntree; i++)
	{
		(unsigned long)trees[i] += (unsigned long)base;
		DIAG((D.s,pid,"tree[%d]>%ld = %lx\n",i,(long)trees[i]-(long)base,trees[i]));

		obj = trees[i];
		if ((obj[3].ob_type & 255) != G_TITLE)		/* Not a menu tree */
			do
			{			/* Fix all object coordinates */
				obj->r.x = (((obj->r.x & 255) * designWidth + (obj->r.x >> 8)) * resWidth) / designWidth;
				obj->r.y = (((obj->r.y & 255) * designHeight + (obj->r.y >> 8)) * resHeight) / designHeight;
				obj->r.w = (((obj->r.w & 255) * designWidth + (obj->r.w >> 8)) * resWidth) / designWidth;
				obj->r.h = (((obj->r.h & 255) * designHeight + (obj->r.h >> 8)) * resHeight) / designHeight;
			} while ( ! (obj++->ob_flags & LASTOB));
		else	/* Standard AES menu */
		{
			j = 0;
			do
				obfix(obj, j);
			while ( ! (obj[j++].ob_flags & LASTOB));
		}
	}
	
	return (void *)base;
}

/* HR */
void Rsrc_setglobal(XA_CLIENT *client)
{
	RSHDR *h = (RSHDR *)client->std_resource;
	OBJECT **o;
	(unsigned long)o = (unsigned long)h + h->rsh_trindex;
/* Fill in the application's global array with a pointer to the resource */
	*((unsigned long *)&client->globl_ptr[5]) = (unsigned long)o;
	*((unsigned long *)&client->globl_ptr[7]) = (unsigned long)h;

	DIAG((D.rsrc,-1,"Resources %ld(%lx) in global[5]\n", o, o));
	DIAG((D.rsrc,-1,"      and %ld(%lx) in global[7]\n", h, h));
}

/*
 * FreeResources : Dispose of a set of loaded resources
 * HR: improvements regarding multiple resources.
 */
void FreeResources(XA_CLIENT *client)
{
	XA_MEMORY *cur = client->memory;
DIAG((D.rsrc,client->pid,"FreeResources: %ld @ %lx for %d, ct=%d\n", cur->size, cur, client->pid, client->rsct));
	if (cur && client->rsct)
	{
		XA_free(client, client->rsct);
		client->rsct--;
/* Put global of older rsrc back, and the pointer in global */
		if (client->memory && client->rsct)
		{
			cur = client->memory;
			while (cur)
			{
				if (cur->id == 1 && cur->ct == client->rsct)		/* Yeah, there is a resource left! */
				{
					client->std_resource = cur->area;
					Rsrc_setglobal(client);
				}
				cur = cur->next;
			}
		}
	}
}


/*
 *	HR: The routines below are almost entirely rewritten such, that it is
 *      quite easy to see the subtle differences.
 *
 * ResourceTree : Find the tree with a given index
 * hdr = pointer to base of resources
 * num = index number of tree
 * Return = pointer to tree or object or stuff, or NULL on failure
 */

#define num_nok(t) (!hdr || num < 0 || num >= hdr->rsh_ ## t)
#define start(t) (unsigned long)index = (unsigned long)hdr + hdr->rsh_ ## t

/* HR: fixing up the pointer array is now done in Loadresources, to make it usable via global[5] */
OBJECT *ResourceTree(RSHDR *hdr, short num)
{
	OBJECT **index;
	
	if num_nok(ntree)
		return nil;
	start(trindex);
	return index[num];
}

/* Find the object with a given index  */
OBJECT *ResourceObject(RSHDR *hdr, short num)
{
	OBJECT *index;

	if num_nok(nobs)
		return nil;
	start(object);
	return index + num;
}

/* Find the tedinfo with a given index */
TEDINFO *ResourceTedinfo(RSHDR *hdr, short num)
{
	TEDINFO *index;

	if num_nok(nted)
		return nil;
	start(tedinfo);
	return index + num;
}

/* Colour icons are too new */

/* Find the iconblk with a given index */
ICONBLK *ResourceIconblk(RSHDR *hdr, short num)
{
	ICONBLK *index;

	if num_nok(nib)
		return nil;
	start(iconblk);
	return index + num;
}

/* Find the bitblk with a given index */
BITBLK *ResourceBitblk(RSHDR *hdr, short num)
{
	BITBLK *index;

	if num_nok(nbb)
		return nil;
	start(bitblk);
	return index + num;
}

/* Find the string with a given index */

/* HR: I think free_strings are the target; The only difference
			is the return value.
	   Well at least TERADESK now works the same as with other AES's */

char *ResourceString(RSHDR *hdr, short num)
{
	char **index;

	if num_nok(nstring)
		return nil;
	start(frstr);
DIAG((D.rsrc,-1,"Gaddr 5 %lx '%s'\n", index[num], index[num]));
	return index[num];
}

/* Find the image with a given index */
/* HR: images (like strings) have no fixed length, so they must be held
       (like strings) in a pointer array, which should have been fixed up.
       Where????? Well I couldnt find it, so I wrote it, see loader.
 */
void *ResourceImage(RSHDR *hdr, short num)
{
	void **index;

	if num_nok(nimages)
		return nil;
	start(frimg);
	return index[num];
}

/* HR: At last the following 2 make sense */

/* Find the ref to the free strings array */
char **ResourceFrstr(RSHDR *hdr, short num)
{
	char **index;

	start(frstr);
DIAG((D.rsrc,-1,"Gaddr 15 %lx '%s'\n", index, *index));
	return index;
}

/* Find the ref to free images array */
void **ResourceFrimg(RSHDR *hdr, short num)
{
	void **index;

	start(frimg);
DIAG((D.rsrc,-1,"Gaddr 16 %lx\n", index));
	return index;
}

AES_function XA_rsrc_load		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	char *n = (char *)(pb->addrin[0]);
	Path full_path;
	short f;

	CONTROL(0,1,1)

/* If the client is overwriting its existing resource then better free it
	 (it shouldn't, but just in case) */
/* I don't think this is a good idea - much better to have a memory leak
	than a process continuing to access a freed memory region; GEM AES
	doesn't auto-free an existing RSC either, so this would be
	incompatible anyway... <mk>

	HR: suggestion: put them in a list, last in front. So you can free what is
	    left in XA_client_exit.
	    7/9/200   done.  As well as the memory allocated for colour icon data.
*/
#if 0
 	if (client->std_resource)
 		FreeResources(client);
#endif

/* What sort of path is it? */
	for (f = 0; (n[f]) && (n[f] != bslash); f++)
		;
	if (n[f])
		sprintf(full_path, "%s", n);
	else
		sprintf(full_path, "%s" sbslash "%s", client->home_path, n);

	DIAG((D.rsrc,client->pid,"rsrc_load('%s')\n", full_path));

	client->std_resource = LoadResources(client, full_path, DU_RSX_CONV, DU_RSY_CONV);
	if (client->std_resource)
	{
		if (client->globl_ptr != pb->globl)
		{
			DIAGS(("WARNING: rsrc_load global is different from appl_init's global\n"));
			
		}
		Rsrc_setglobal(client);
		pb->intout[0] = 1;
	}
	else
		pb->intout[0] = 0;

	return XAC_DONE;
}

AES_function XA_rsrc_free		/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(0,1,0)

	if (client->std_resource)
	{
		FreeResources(client);
		pb->intout[0] = 1;
	} else
	{
		pb->intout[0] = 0;
	}

	return XAC_DONE;
}

AES_function XA_rsrc_gaddr	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	void **addr = (void **)pb->addrout;
	void **glob;
	short type = pb->intin[0], index = pb->intin[1];

	CONTROL(2,1,0)

	if (!client->std_resource)
	{
		glob = (void **)(client->globl_ptr + 7);
		client->std_resource = *glob;
	}

	DIAG((D.rsrc,-1,"rsrc_gaddr type %d, index %d\n", type, index));

	if (addr)	/* HR: It is better to let the client bomb out in stead of precious us. */
	{
		pb->intout[0] = 1;

		switch(type)
		{
		case R_TREE:
			*addr = ResourceTree(client->std_resource, index);
			break;
		case R_OBJECT:
			*addr = ResourceObject(client->std_resource, index);
			break;
		case R_TEDINFO:
			*addr = ResourceTedinfo(client->std_resource, index);
			break;
		case R_ICONBLK:
			*addr = ResourceIconblk(client->std_resource, index);
			break;
		case R_BITBLK:
			*addr = ResourceBitblk(client->std_resource, index);
			break;
/*!*/	case R_STRING:
			*addr = ResourceString(client->std_resource, index);		/* HR */
			break;
/*!*/	case R_IMAGEDATA:
			*addr = ResourceImage(client->std_resource, index);
			break;
		case R_OBSPEC:
			*addr = (void *)ResourceObject(client->std_resource, index)->ob_spec;
			break;
		case R_TEPTEXT:
			*addr = ResourceTedinfo(client->std_resource, index)->te_ptext;
			break;
		case R_TEPTMPLT:
			*addr = ResourceTedinfo(client->std_resource, index)->te_ptmplt;
			break;
		case R_TEPVALID:
			*addr = ResourceTedinfo(client->std_resource, index)->te_pvalid;
			break;
		case R_IBPMASK:
			*addr = ResourceIconblk(client->std_resource, index)->ib_pmask;
			break;
		case R_IBPDATA:
			*addr = ResourceIconblk(client->std_resource, index)->ib_pdata;
			break;
		case R_IBPTEXT:
			*addr = ResourceIconblk(client->std_resource, index)->ib_ptext;
			break;
		case R_BIPDATA:
			*addr = ResourceBitblk(client->std_resource, index)->bi_pdata;
			break;
/*!*/	case R_FRSTR:
			*addr = ResourceFrstr(client->std_resource, index);
			break;
/*!*/	case R_FRIMG:
			*addr = ResourceFrimg(client->std_resource, index);
			break;
		}
	}	
	else
		pb->intout[0] = 0;
	return XAC_DONE;
}

AES_function XA_rsrc_obfix	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(1,1,1)

	obfix((OBJECT *)pb->addrin[0], pb->intin[0]);
	pb->intout[0] = 1;
	return XAC_DONE;
}

