/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#ifndef _XA_TYPES_H_
#define _XA_TYPES_H_
#include "k_defs.h"
#include "xa_defs.h"
/* #include "hr_mem.h"	*/	/* Thread save memory allocator for client memory. */

#define MONO (display.colours < 16)		/* HR */

/*-----------------------------------------------------------------
 * Lock control
 *-----------------------------------------------------------------*/

typedef enum locks
{
	NOLOCKS   = 0,
	appl      = 1,
	newclient = 2,
	trap      = 4,
	winlist   = 010,
	desk     = 011,
	clients = 012,
	fsel   = 014,
	screen = 020,
	mouse  = 021,
	envstr = 022,
	NOLOCKING = -1
} LOCK;

/*		Not yet :-)
typedef struct process_tracking
{
	LOCK lock;
	short pid;
} TRACK;
*/

#include "messages.h"

/*-----------------------------------------------------------------
 * Display descriptions
 *-----------------------------------------------------------------*/

typedef enum {D_LOCAL = 0, D_TCP = 1} XA_DISPLAY;

typedef struct {
	short bg_col;			/* Colour used for backgrounds */
	short fg_col;			/* Colour used for foregrounds */
	short b_r_col;			/* Colour used for bottom right 3d edge */
	short t_l_col;			/* Colour used for top left 3d edge */
	short border_col;		/* Colour used for edging */
	short highlight_col;	/* Colour used for highlighting */
} XA_COLOUR_SCHEME;

typedef struct {
	RECT r;						/* Screen dimensions */
	short colours;				/* Number of colours available */
	short planes;				/* Number of planes in screen */
	XA_DISPLAY display_type;	/* The type of display we are using */
	short P_handle;				/* VDI physical workstation handle (for D_LOCAL screens) */
	short V_handle;				/* VDI virtual workstation handle (for all screens) */
	short c_max_w, c_max_h;		/* Maximum character dimensions in pixels */
	short c_min_w, c_min_h;		/* Minimum (small font) character dimensions in pixels */
	XA_COLOUR_SCHEME dial_colours;	/* Colour scheme used for dialogs */
	short standard_font_height;	/* Needed for appl_getinfo */
	short standard_font_id;
	short standard_font_point;
	short small_font_id;
	short small_font_point;
	short small_font_height;	/* Needed for appl_getinfo */
} XA_SCREEN;

/*-----------------------------------------------------------------
 * Object Handler Structures
 *-----------------------------------------------------------------*/

/*
 * Object display callback parameter
 * - This only exists because Lattice 5.52 has a bug that crashes the compiler
 *   if you have:
 *		typedef void (*ObjectDisplay)(OBJECT *ob, short parent_x, short parent_y);
 * HR: (*)()    I removed the (*...) out of ALL typedef'd prototypes. :-)
 */
/* HR: Well, it appeared to be handy as well. */
typedef struct {
	OBJECT *tree;
	short object;
	RECT r;				/* HR: Why not do the addition (parent_x+ob_x) once in the caller? */
						/* HR: And set a usefull RECT as soon as possible, ready for use in
								all kind of functions. */
	short parent_x;		/* Keep both in: dont need to change everything in a single effort */
	short parent_y;
	short *state_mask;
} ODC_PARM;

/* Object display function type */
typedef void ObjectDisplay(ODC_PARM *odc_p);	/* HR (*)() */
/* Object handler function type */
typedef void ObjectHandler(LOCK lock, ODC_PARM *odc_p); /* HR */

/*-----------------------------------------------------------------
 * Client application data structures
 *-----------------------------------------------------------------*/

/* List of Pending AES messages */
typedef struct xa_aesmsg_list
{
	struct xa_aesmsg_list *next;
	MSG_BUF message;
} XA_AESMSG_LIST;

/* HR: Keystrokes must be put in a queue if no apps are waiting yet.
   There also may be an error in the timer event code.
   A queue like this is pretty anyhow, even if it is almost never needed.
*/
typedef struct key_queue
{
	short cur,end;
	short k[64];
} KEY_Q;

typedef struct
{
	short mx;
	short my;
	short state;
	short clicks;
} MOUSE_DATA;

typedef struct button_button
{
	struct xa_client *client;
	short state;
} BUT_Q;

/*
 * Codes for what types of event an app is currently waiting for
 * - I've used the same bit allocation as evnt_multi() here, but there is
 *   the extra XAWAIT_MULTI bit to indicate we are waiting from evnt_multi()
 *   as opposed to one of the individual routines (as return values are different).
 */
typedef enum
{
	XAWAIT_KEY		= 0x01,
	XAWAIT_BUTTON	= 0x02,
	XAWAIT_MOUSE1	= 0x04,
	XAWAIT_MOUSE2	= 0x08,
	XAWAIT_MESSAGE	= 0x10,
	XAWAIT_TIMER	= 0x20,
	XAWAIT_MULTI	= 0x100,	/* Waiting for an evnt_multi() */
	XAWAIT_DIALOG	= 0x200,	/* Waiting for a form_do() dialog to exit */
	XAWAIT_CHILD	= 0x400,	/* Waiting for a non-parallel shel_write() child process to exit */
/* HR: menu rectangle event */
	XAWAIT_MENU		= 0x1000,	/* AES private */
} WAITING_FOR;

struct xa_window;
struct xa_widget;

/* Client Types HR: change to names from compendium (some time :-) */
#define XA_CT_APP	0x01	/* Client is a standard app */
#define XA_CT_ACC	0x02	/* Client is an accessory */

typedef enum		/* HR: compendium */
{
	APP_FIRST,
	APP_NEXT,
	APP_DESK,
	APP_SYSTEM      = 0x01,
	APP_APPLICATION = 0x02,
	APP_ACCESSORY   = 0x04,
	APP_SHELL       = 0x08
} CLIENT_TYPE;

typedef void TASK(struct task_administration_block *tab);

typedef struct
{
	short flags;
	RECT m1,m2;
	TASK *t1,*t2;	/* used by tasking */
} XA_MOUSE_RECT;

#define ATTACH_MAX 64			/* This is per menu_bar!!!! */

typedef struct menu_attachments	/* HR */
{
	char *text;		/* outplaced menu entry text, to which a left arrow is appended */
	short item;
/*	short pid;
*/	OBJECT *sm;		/* attached menu tree */
} XA_MENU_ATTACHMENT;

/* Object Tree based widget descriptor */
typedef struct widget_tree
{
	OBJECT *tree;				/* The object tree */
	LOCK lock;
	short edit_obj,				/* Index of the current editable text field (if any) */
	      edit_pos,				/* Cursor position within the text field (if any) */
	      owner,				/* The tree widget would be owned by a different app to
								   the actual window (like the menu bar on the root window) */
	      lastob;				/* Can be used to validate item number */
	XA_MENU_ATTACHMENT *at;		/* extended ob_spec's */
	ObjectHandler *handler;		/* Handler vec for direct handling */
} XA_WIDGET_TREE;

/* used for linking Mxalloc'd area's per client */
typedef struct xa_memory
{
	struct xa_memory *next, *prior;
	short id, ct;
	long size;
	char area[0];
} XA_MEMORY;

/* Note that I dont use PATH_MAX here; this is only concerning
   allocation of enough space throughout. */
typedef char Path[132];			/* HR */

/* Main client application descriptor */
typedef struct xa_client
{
	struct xa_client *next, *prior;
	bool init;				/* HR: true between appl_init & appl_exit */
	short pid;				/* HR: clients process identifier number */
	int client_end,			/* File handle's of the clients internal reply pipe */
	    kernel_end;
	XA_AESMSG_LIST *msg;	/* Pending AES messages */
	struct xa_pending_widget *widget_active;	/* Pending active widget (if any) */
	WAITING_FOR waiting_for;		/* What types of event(s) the client is waiting for */
	AESPB *waiting_pb;		/* Parameter block for whatever the client is waiting for */
	short client_mouse;		/* The cursor to use when this is top application */
	MFORM *client_mouse_form;
	short *globl_ptr;		/* Pointer to the client's globl array (so we can fill in the resource
							   address field later). */
	void *std_resource;		/* Pointer to the client's standard GEM-style single resource file */
	short rsct;				/* HR: count up/down the loaded resources. Used by XA_alloc, XA_free */
/* HR: */
	XA_WIDGET_TREE std_menu;/* The client's standard GEM-style menu-bar widget */
	Path home_path;			/* The directory that the client was started in */
	char *cmd_tail;			/* The command tail of the process (if launched by shell_write) */
	bool heap_tail;			/* If true, cmd_tail is a malloc */
	char *cmd_name;			/* The full filename used when launching the process (if launched by shell_write) */
	bool heap_name;			/* If true, cmd_name is a malloc */
	char name[NICE_NAME];	/* The clients 'pretty' name (possibly set by menu_register) */
	char proc_name[9];		/* The clients 'official' (ie. used by appl_find) name. */
	short parent;			/* Process id of the client's "parent" process (if launched with shell_write) */
	struct xa_window *zen;	/* Pointer to a window that could be about to become a dialog */
	void *temp;				/* Temporary storage */
	OBJECT *desktop;		/* Pointer to the client's custom desktop object tree (if any) */
	short type;				/* What type of client is this? */
	bool half_screen;		/* HR: true if the client has issued wind_get(WF_SCREEN, ...) */
	XA_MOUSE_RECT em;		/* HR: Needed. whether threads or not */
	XA_MEMORY *memory;		/* HR: link loaded resoures' allocated memory, so it can be freeed.
	                               this also solves the problem that memory allocated for colour icon data
	                               was left orphaned. */
} XA_CLIENT;

/* HR:  removing the (*) makes it more usable */
typedef
unsigned long AES_function(LOCK lock, XA_CLIENT *client, AESPB *pb);		/* HR: pass the client structure. */

/*-----------------------------------------------------------------
 * Windows & Widgets
 *----------------------------------------------------------------- */

/* HR (*)() */
/* A function of the type used for widget behaviours is a 
   'WidgetBehaviour'. */
typedef short WidgetBehaviour(LOCK lock, struct xa_window *wind, struct xa_widget *widg);

/* Relative coordinate types */
typedef enum
{
	RT,			/* Top right */
	RB,			/* Bottom right */
	LT,			/* Top left */
	LB,			/* Bottom left */
	CT,			/* Top centred */
	CR,			/* right centered */
	CB,			/* Bottom centred */
	CL			/* Left centered */
} XA_RELATIVE;

/* Widget positions are defined as relative locations */
typedef struct
{
	XA_RELATIVE relative_type;	/* Where is the position relative to? */
	RECT r;						/* Position */
} XA_WIDGET_LOCATION;

/* Widget Callback/Behaviour Identifiers */
typedef enum
{
	XACB_DISPLAY,
	XACB_CLICK,
	XACB_DCLICK,
	XACB_DRAG,
	COUNT_XACB
} XACB;

/* Widget Index's */

typedef enum
{
	XAW_TITLE,
	XAW_CLOSE,
	XAW_FULL,
	XAW_MOVER,			/* Not actually used like the others */
	XAW_INFO,
	XAW_RESIZE,
	XAW_UPLN,
	XAW_DNLN,
	XAW_VSLIDE,
#if NEW_SLIDE		/* HR */
	XAW_UPPAGE,	
	XAW_DNPAGE,
#endif
	XAW_LFLN,
	XAW_RTLN,
	XAW_HSLIDE,
#if NEW_SLIDE		/* HR */
	XAW_LFPAGE,
	XAW_RTPAGE,
#endif
	XAW_ICONIFY,
	XAW_MENU,			/* Extended XaAES widget */
	XAW_TOOLBAR,		/* Extended XaAES widget */
/* Number of available XA_WIDGET slots in a the window for default/standard widgets */
	XA_MAX_WIDGETS
} XA_WIDGETS;

typedef enum
{
	PULL,
	PUSH,
	LEAVE
} MENU_BEHAVE;

#if 0   /* HR: use ob_state bits */
/* Widget Status */
typedef enum {XAW_PLAIN = 0, XAW_SELECTED = 1, XAW_TOOLBARSTATE = 3} XA_WIDGET_STATUS;
#endif

/* Window Widget */
typedef struct xa_widget
{
	XA_WIDGET_LOCATION loc;			/* Location of widget relative to window extents */
	WidgetBehaviour *behaviour[COUNT_XACB];	/* Callback function pointers to the behaviours of the widget */
	OB_STATES stat;					/* Current status (selected, etc) */
	XA_WIDGETS type;				/* HR: For convenience, makes it possible to dusconnect type from index */
	short mask;						/* HR: disconnect NAME|SMALLER etc from emumerated type XA_WIDGETS */
	short click_x, click_y;			/* If we are displaying because we've been clicked on, this is the location  */
									/* of the click (this is only used by the slider widget really) */
	short rsc_index,			/* HR: If a bitmap widget, index in rsc file */
		  arrowx,				/* HR: WM_ARROWED msg type */
		  limit,				/* HR: on which limit to stop */ 
		  slider_type;			/* HR: which slider should move for scroll widget */
	void *stuff;				/* Pointer to widget dependant context data, if any */
	short start;				/* HR: If stuff is a OBJECT tree, we want start drawing here */
} XA_WIDGET;

typedef struct xa_widget_list
{
	XA_WIDGET this;				/* ugly!! :-) */
	struct xa_widget_list *next;
} XA_WIDGET_LIST;

/* Pending action from a widget */
typedef struct xa_pending_widget
{
	XA_WIDGET *widg;				/* Pointer to widget for which this action is pending */
	struct xa_window *wind;			/* Window to which the widget is connected */
	WidgetBehaviour *action;			/* Callback function to call */
	short info;						/* Action specific information */
	short info2;
	short info3;
} XA_PENDING_WIDGET;

/* Extra details for a slider widget */

typedef struct
{
	short position;		/* Actual position of the slider (0-1000) */
	short length;		/* Length (0-1000) */
	RECT r;				/* physical */
} XA_SLIDER_WIDGET;

#define ZT_A	1
#define ZT_B	2
#define ZT_C	3
#define ZT_D	4

/* Rectangle List entry */
typedef struct xa_rect_list
{
	struct xa_rect_list *next;
	RECT r;			/* Dimensions of segment */
} XA_RECT_LIST;

/* Callback for a window's auto-redraw function */
/* HR: (*)() */
typedef short WindowDisplay (LOCK lock, struct xa_window *wind);
typedef short WindowKeypress (LOCK lock, struct xa_window *wind, unsigned short keycode);

/* Window status codes */
typedef enum
{
	XAWS_CLOSED,
	XAWS_OPEN,
	XAWS_ICONIFIED
} XAWS;

/* Window Descriptor */
typedef struct xa_window
{
	struct xa_window *next,	/* Window list stuff	- next is the window below */
	                 *prev;	/*						- prev is the window above (both NULL terminated) */
	long active_widgets;	/* Summary of the current standard widgets for the window */
	bool nolist;			/* If true, dont put in the window_list. For embedded listboxes mainly. */
	RECT max,				/* Creator dimension's, maximum for sizing */
	     r,					/* Current dimensions */
		 wa,				/* user work area */
		 bd,				/* border displacement */
		 pr;				/* previous dimensions */
	RECT *remember;			/* Where to store the current position for remembering. */
	short handle,			/* Window handle */
	      owner,			/* Client pid of the window's owner */
	      behalf,			/* HR: pid on behalf of which a window is created: dialogues, alerts, and fileselector. */
	      is_open,			/* Flag - is the window actually open? */
	      window_status,	/* Window status */
	      frame,			/* HR: Size of the frame (0 for windowed listboxes) */
	      widgets_on_top;	/* Number of widgets on the top bar, not including the title */
	XA_RECT_LIST *rect_list,	/* The rectangle list for redraws in this window */
	             *rect_user,	/* User (wind_get) rect list current pointer */
	             *rect_start;	/* Start of the rectangle list memory block */
	void *background;		/* Pointer to a buffer containing the saved background */
	short bgx, bgy;			/* Location the saved background came from */
	short created_by_FMD_START;	/* Flag - 1 = this window was created by form_dial, */
								/* so don't destroy it until the form_dial(FMD_FINISH) */
	WindowDisplay*redraw;	/* Pointer to the window's auto-redraw function (if any) */
	WindowKeypress *keypress;	/* Pointer to the window's keyboard handler function (if any) */
	WindowDisplay  *destructor;	/* Pointer to the window's destructor function (auto-called on window deletion if a fully auto window) */
	SendMessage *send_message;	/* Pointer to the internal message handler for this window (to be handled by the creator) */
	OBJECT *winob;			/* HR: Tree and index of a sub window (The parent object of the window) */
	short item;				/*      currently used by list boxes */
#if USER_WIDGETS
	XA_WIDGET_LIST *user_widgets;		/* Pointer to a list of user added widgets */
#endif
	XA_WIDGET widgets[XA_MAX_WIDGETS];	/* The windows standard widget set (array for speed) */
	XA_PENDING_WIDGET *widget_active;	/* Pending active widget (if any) */
} XA_WINDOW;


typedef struct win_base			/* HR: open and closed windows in separate lists. */
{
	XA_WINDOW *first,*last;
} WIN_BASE;

#define root_window open_windows.last
#define window_list open_windows.first

/* HR (*)()    prefer '*' outside typedef */
typedef void fsel_handler(LOCK lock, char *path, char *file);		/* HR */
typedef short scrl_click(LOCK lock, OBJECT *form, short objc);		/* HR */

/* HR:  The FS_LIST box is the place holder and the
        entrypoint via its TOUCHEXIT flag.
        The list box itself is turned into a full fledged window.
*/

/* Directory entry flags */
typedef enum scroll_entry_type
{
	FLAG_DIR        = 1,
	FLAG_EXECUTABLE = 2,
	FLAG_LINK       = 4,
	FLAG_MAL        = 0x100,	/* text part is malloc'd and must be freed. */
	FLAG_ENV        = 0x200
} SCROLL_ENTRY_TYPE;

typedef struct scroll_entry
{
	char *text;					/* Text to display */
	OBJECT *icon;				/* Icon/object to display to the left of text (if any) */
	struct scroll_entry *next;	/* Next element */
	struct scroll_entry *prev;	/* Previous element */
/* This simple number makes everything so easy */
	short n;					/* HR: element number in the list */
	SCROLL_ENTRY_TYPE
	    flag;					/* HR */
	char the_text[0];			/* if text included in malloc it is here.
									Let entry->text point to here.
									FLAG_MAL is off */
} SCROLL_ENTRY;

/* HR: please NO local prefixes! */

typedef void scrl_widget(struct scroll_info *list);

typedef struct scroll_info
{
	LOCK lock;
	XA_WINDOW *wi,				/* HR: make the scroll box a real window */
	          *pw;				/* HR: If the listbox is part of a windowed dialogue, we must know that,
	                                   otherwise we couldnt move that window (rp_2_ap). */
	OBJECT *tree;				/* HR: originating object */
	short item;
	SCROLL_ENTRY
		*start,					/* Pointer to first element */
		*cur,					/*            current selected element */
		*top,					/*            top-most displayed element */
		*bot,					/*            bottom_most displayed element */
		*last;					/*            last element */
	char *title;				/* Title of the list */
	/* The following simple numbers make everything so easy */
	short n,					/* Number of elements in the list */
		  s,					/* HR: Number of elements that fits in the box */
	      v,					/* HR: number of characters that fit the width of the box */
	      left,					/* HR: Portion of the line that is left of the visible part */
	      max,					/* HR: Maximum length of line allowed */
	      state;				/* Extended status info for scroll list */
	scrl_click *dclick,			/* HR Callback function for double click behaviour */
	           *click;			/* HR Callback function for single click behaviour */
	scrl_widget *slider,		/* HR slider calc function */
	            *closer,		/* HR closer function */
	            *fuller;		/* HR fuller function */
} SCROLL_INFO;

/* SCROLL_INFO .status bits */
#define SCRLSTAT_UP		1		/* Scroll up is clicked on */
#define SCRLSTAT_DOWN	2		/* Scroll down is clicked on */
#define SCRLSTAT_SCROLL	4		/* Scroller drag is clicked on */
#define SCRLSTAT_RDB	8		/* Redraw buttons only */

#endif
