/*
 * XaAES - XaAES Ain't the AES
 *
 * A multitasking AES replacement for MiNT
 *
 */

#include <mintbind.h>
#include <osbind.h>
#include <memory.h>

#include "xa_types.h"
#include "xa_globl.h"

#include "c_window.h"
#include "rectlist.h"
#include RSCHNAME
#include "objects.h"
#include "widgets.h"
#include "desktop.h"
#include "xa_form.h"
#include "xa_graf.h"
#include "xa_rsrc.h"
#include "app_man.h"

/*
 *	AES window handling functions
 *	I've tried to support some of the AES4 extensions here as well as the plain
 *	single tasking GEM ones.
 */

AES_function XA_wind_create	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_WINDOW *new_window;
	RECT *r = (RECT *)&pb->intin[1];				/* HR */

	CONTROL(5,1,0)	

	new_window = create_window(lock, send_app_message, client->pid, false, pb->intin[0], MG, *r, r, nil);		/* HR */

	if (new_window)
		pb->intout[0] = new_window->handle;	/* Return the window handle in intout[0] */
	else
		pb->intout[0] = -1;			/* Fail to create window, return -ve number */

	return XAC_DONE;				/* Allow the kernal to wake up the client - we've done our bit */
}

/* HR: open_window: internal function */
AES_function XA_wind_open	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(5,1,0)	

	pb->intout[0] = open_window(lock, client->pid, pb->intin[0], *(RECT *)&pb->intin[1]);
	return XAC_DONE;
}

AES_function XA_wind_close	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_WINDOW *w;
	
	CONTROL(1,1,0)	

	DIAGS(("XA_wind_close\n"));
	w = get_wind_by_handle(lock, pb->intin[0]);		/* Get the window */

	if (w == NULL)
	{
		DIAGS(("WARNING:wind_close:Invalid window handle\n"));
		pb->intout[0] = 0;			/* Invalid window handle, return error */
		return XAC_DONE;
	}

	if (w->owner != client->pid)	/* Clients can only close their own windows */
	{
		DIAGS(("WARNING: clnt %d cannot close window %d (not owner)\n", client->pid, w->handle));
		pb->intout[0] = 0;			/* Invalid window handle, return error */
		return XAC_DONE;
	}
	
	pb->intout[0] = close_window(lock, w);
	
	return XAC_DONE;
}

AES_function XA_wind_find		/* (LOCK lock, XA_CLIENT *client,AESPB *pb) */
{
	XA_WINDOW *w = wind_find(lock, pb->intin[0], pb->intin[1]);	/* Is there a window under the mouse? */

	CONTROL(2,1,0)	

	pb->intout[0] = w ? w->handle : 0;

	return XAC_DONE;
}

/* HR: background now integrated in display slider */
static
void update_vslide(LOCK lock, XA_WINDOW *wind)
{
	XA_RECT_LIST *rl;
	XA_WIDGET *widg = wind->widgets + XAW_VSLIDE;
	
	v_hide_c(V_handle);
	for (rl = rect_get_system_first(wind); rl; rl = rect_get_system_next(wind))
	{		/* Walk the rectangle list */
		set_clip(&rl->r);
		display_vslide(lock, wind, widg);
	}
	v_show_c(V_handle, 1);
}

static
void update_hslide(LOCK lock, XA_WINDOW *wind)
{
	XA_RECT_LIST *rl;
	XA_WIDGET *widg = wind->widgets + XAW_HSLIDE;
	
	v_hide_c(V_handle);
	for (rl = rect_get_system_first(wind); rl; rl = rect_get_system_next(wind))
	{		/* Walk the rectangle list */
		set_clip(&rl->r);
		display_hslide(lock, wind, widg);
	}
	v_show_c(V_handle, 1);
}

void top_window(LOCK lock, XA_WINDOW *w, short menu_owner)			/* HR */
{
	short pid = w->owner;
	XA_WINDOW *old_focus, *new_focus;
	XA_CLIENT *client = Pid2Client(pid);

	DIAG((D.wind,-1,"top_window %d for %s(%d)\n",  w->handle, client->name, pid));

	old_focus = window_list;

	pull_wind_to_top(lock, w);	/* New top window - change the cursor to this client's choice */

	focus = w->owner != menu_owner ? root_window : window_list;		/* HR */
	
	graf_mouse(client->client_mouse, client->client_mouse_form);
	display_non_topped_window(lock, 40, old_focus, NULL);  /* redisplay title */
	if (old_focus != w)
		display_non_topped_window(lock, 41, w, NULL);		/* Display the window */

	w->send_message(lock, w, w->owner, WM_REDRAW, 0, w->handle, w->r.x, w->r.y, w->r.w, w->r.h);
}

void bottom_window(LOCK lock, XA_WINDOW *w)
{
	short was_top = (w == window_list);
	RECT clip;
	XA_WINDOW *wl = w->next;

DIAG((D.wind,-1,"botom_window %d\n", w->handle));

	if (wl == root_window)			/* Don't do anything if already bottom */
		return;

	send_wind_to_bottom(lock, w);			/* Send it to the back */

DIAG((D.wind,-1," - menu_owner %d, w_list_owner %d\n", menu_owner(), window_list->owner));
	focus = window_list->owner != menu_owner() ? root_window : window_list;		/* HR */

	display_non_topped_window(lock, 42, w, NULL);			/* Redisplay titles */
	display_non_topped_window(lock, 43, window_list, NULL);

	while (wl != w)			/* Our window is now right above root_window */
	{
		clip = wl->r;
		if (rc_intersect(&w->r, &clip))
		{
			display_non_topped_window(lock, 44, wl, &clip);		/* Re-display any revealed windows */
			if ( ! (wl->active_widgets & NO_MESSAGES))
				wl->send_message(lock, wl, wl->owner, WM_REDRAW, 0, wl->handle, clip.x, clip.y, clip.w, clip.h);
		}
		wl = wl->next;
	}

	if (was_top && !(window_list->active_widgets & NO_MESSAGES))
		window_list->send_message(lock, window_list, window_list->owner, WM_ONTOP, 0, window_list->handle, 0, 0, 0, 0);
}

AES_function XA_wind_set	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_WINDOW *w;
	OBJECT *ob;
	RECT clip;
	short wind = pb->intin[0], cmd = pb->intin[1];
	unsigned short *l;
	char *t;

	CONTROL(6,1,0)	

	w = get_wind_by_handle(lock, wind);

	DIAG((D.wind,client->pid,"wind_set: w%lx, h%d, cmd%d\n",w,w ? w->handle : -1, cmd));

	if (w == NULL)
	{
		switch(cmd)
		{
		case WF_NEWDESK:			/* Some functions may be allowed without a real window handle */
			break;
		default:
			DIAGS(("WARNING:wind_set:Invalid window handle =%d\n", w));
			pb->intout[0] = 0;		/* Invalid window handle, return error */
			return XAC_DONE;
		}
	} else
		if ((w->owner != client->pid)		/* Clients can only change their own windows */
			&& ((w != root_window) || (cmd != WF_NEWDESK)))
		{
			DIAGS(("WARNING: clnt %d cannot change window %d (not owner)\n", client->pid, w->handle));
			pb->intout[0] = 0;		/* Invalid window handle, return error */
			return XAC_DONE;
		}
	
	switch(cmd) {
	case WF_HSLIDE:
		if (w->widgets[XAW_HSLIDE].stuff)
		{
			XA_SLIDER_WIDGET *slw = w->widgets[XAW_HSLIDE].stuff;
			slw->position = bound_sl(pb->intin[2]);		/* HR */
			update_hslide(lock, w);
		}
		break;
	case WF_VSLIDE:
		if (w->widgets[XAW_VSLIDE].stuff)
		{
			XA_SLIDER_WIDGET *slw = w->widgets[XAW_VSLIDE].stuff;
			slw->position = bound_sl(pb->intin[2]);		/* HR */
			update_vslide(lock, w);
		}
		break;
	case WF_HSLSIZE:
		if (w->widgets[XAW_HSLIDE].stuff)
		{
			XA_SLIDER_WIDGET *slw = w->widgets[XAW_HSLIDE].stuff;
			slw->length = bound_sl(pb->intin[2]);		/* HR */
			update_hslide(lock, w);
		}
		break;
	case WF_VSLSIZE:
		if (w->widgets[XAW_VSLIDE].stuff)
		{
			XA_SLIDER_WIDGET *slw = w->widgets[XAW_VSLIDE].stuff;
			slw->length = bound_sl(pb->intin[2]);		/* HR */
			update_vslide(lock, w);
		}
		break;
	case WF_NAME:
		l = (unsigned short *)(pb->intin);
		t = (char*)((long)l[2] << 16);
		t += l[3];
		if (t == NULL)			/* HR */
			t = "";
		w->widgets[XAW_TITLE].stuff = t;

		if (w->is_open)
		{
			rp_2_ap(w, w->widgets + XAW_TITLE, &clip);
			display_non_topped_window(lock, 45, w, &clip);
		}
		break;
	case WF_INFO:
		l = (unsigned short *)(pb->intin);
		t = (char *)((long)l[2] << 16);
		t += l[3];
		if (t == NULL)			/* HR */
			t = "";
		w->widgets[XAW_INFO].stuff = t;

		if ((w->active_widgets & INFO) && (w->is_open))
		{
			rp_2_ap(w, w->widgets+XAW_INFO, &clip);
			display_non_topped_window(lock, 46, w, &clip);
		}
		break;

	case WF_CURRXYWH:				/* Move a window, HR: check sizes */
		{
			short mw = pb->intin[4],
			      mh = pb->intin[5];
			if (w->max.w && mw > w->max.w)
				mw = w->max.w;
			if (w->max.h && mh > w->max.h)
				mh = w->max.h;
			move_window(lock, w, pb->intin[2], pb->intin[3], mw, mh);
		}
		break;
	case WF_BEVENT:					/* HR */
		if (pb->intin[2]&1)
			w->active_widgets |= NO_TOPPED;
		else
			w->active_widgets &= ~NO_TOPPED;
		break;
	case WF_BOTTOM:					/* Extension, send window to the bottom */
		if (w->is_open)
			bottom_window(lock, w);
		break;
			
	case WF_TOP:					/* Top the window */
		if (!w->is_open)
			break;
		if (    w != window_list
			|| (w == window_list && focus != w)		/* HR */
			)
		{
			top_window(lock, w, w->owner);
			swap_menu(lock, w->owner);		/* HR: needed because the changed behaviour in close_window. */
		}
		break;
			
	case WF_NEWDESK:				/* Set a new desktop object tree */
		l = (unsigned short *)pb->intin;
		t = (char *)((long)l[2] << 16);
		t += l[3];
		ob = (OBJECT *)t;
			
		if (ob)
		{
			client->desktop = ob;
			set_desktop(ob);
			root_window->owner = client->pid;
		} else
		{
			client->desktop = NULL;
			/* HR shouldnt we find another app that has a client->desktop ? */
			set_desktop(ResourceTree(system_resources, DEF_DESKTOP));
			display_non_topped_window(lock, 47, root_window, NULL);
			root_window->owner = AESpid;				/* HR */
		}
#if 0
		display_non_topped_window(lock, root_window, NULL);
#endif		
		break;
			
	case WF_AUTOREDRAW:				/* Set an auto-redraw callback function for the window */
		l = (unsigned short *)pb->intin;
		t = (char *)((long)l[2] << 16);
		t += l[3];
		w->redraw = (WindowDisplay *)t;		/* HR: * */
		break;
			
	case WF_STOREBACK:				/* Set the window's 'preserve own background' attribute */
#if 0
		w->background = malloc(calc_back(&w->r,display.planes));		/* HR: never used */
#endif
		w->bgx = -1;
		w->active_widgets |= STORE_BACK|NO_REDRAWS;
		break;

/* HR Psema.... already in move_window */
	case WF_ICONIFY:				/* Iconify a window */
		w->window_status = XAWS_ICONIFIED;		/* HR: set before redisplaying :-) */
		move_window(lock, w, pb->intin[2], pb->intin[3], pb->intin[4], pb->intin[5]);
		break;

	case WF_UNICONIFY:				/* Un-Iconify a window */
		w->window_status = XAWS_OPEN;			/* HR: set before redisplaying :-) */
		move_window(lock, w, w->pr.x, w->pr.y, w->pr.w, w->pr.h);
		break;

	}

	pb->intout[0] = 1;
	return XAC_DONE;
}

/*
 * AES wind_get() function.
 * This includes support for most of the AES4 / AES4.1 extensions,
 * with the exception of WF_BEVENT (all XaAES windows get button events in the
 * background at the moment).
 */

AES_function XA_wind_get	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_WINDOW *w;
	XA_RECT_LIST *rl;
	XA_SLIDER_WIDGET *slw;
	short *o = pb->intout;
	short wind = pb->intin[0], cmd = pb->intin[1];
	RECT s, d;
	
	CONTROL(2,5,0)	

	w = get_wind_by_handle(lock, wind);

DIAG((D.wind,client->pid,"wind_get: pid=%d, w=%lx, h:%d, cmd=%d\n",client->pid,w,wind,cmd));
	if (w == NULL)
	{
		switch(cmd)
		{
		case WF_FULLXYWH:			/* Some functions may be allowed without a real window handle */
		case WF_BOTTOM:
		case WF_TOP:
		case WF_NEWDESK:
		case WF_SCREEN:				/* HR */
#ifdef __GNUC__
		case 0x5841:
#else
		case 'XA':						/* HR: idee stolen from WINX */
#endif
			break;
		default:
			DIAGS(("WARNING:wind_get:Invalid window handle =%d\n", wind));
			o[0] = 0;		/* Invalid window handle, return error */
			return XAC_DONE;
		}
	}

	o[0] = 1;

	switch(cmd)
	{
	case WF_FIRSTXYWH:				/* Generate a rectangle list and return the first entry */
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
		rl = rect_get_user_first(w);
/* HR: Oh, Oh  Leaving a intersect unchecked!! */
		if (rl)
		{
			d = w->wa;
			if ( ! rc_intersect(&rl->r, &d))
				goto next;		
			o[1] = d.x;		/* Return the first rectangle coords */
			o[2] = d.y;
			o[3] = d.w;
			o[4] = d.h;
		} else
		{
			o[1] = w->r.x;		/* Totally obscured window, return w & h as 0 */
			o[2] = w->r.y;
			o[3] = 0;
			o[4] = 0;
		}
DIAG((D.wind,-1,"first for %s %d/%d,%d/%d\n", client->name, o[1], o[2], o[3], o[4]));
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
		break;
			
	case WF_NEXTXYWH:				/* Get next entry from a rectangle list */
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
next:
		do{
			rl = rect_get_user_next(w);
			if (rl)
			{
				d = w->wa;
				if (rc_intersect(&rl->r, &d))
				{
					o[1] = d.x;		/* Return the next rectangle coords */
					o[2] = d.y;
					o[3] = d.w;
					o[4] = d.h;
					break;
				}
			} else
			{
				o[1] = w->r.x;		/* Totally obscured window, return w & h as 0 */
				o[2] = w->r.y;
				o[3] = 0;
				o[4] = 0;
				break;
			}
		}od
DIAG((D.wind,-1,"next for %s %d/%d,%d/%d\n", client->name, o[1], o[2], o[3], o[4]));
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
		break;
			
	case WF_CURRXYWH:				/* Get the current coords of the window */
		o[1] = w->r.x;			/* Return the window coords */
		o[2] = w->r.y;
		o[3] = w->r.w;
		o[4] = w->r.h;
		break;
			
	case WF_WORKXYWH:				/* Get the current coords of the window's user work area */
		o[1] = w->wa.x;	
		o[3] = w->wa.w;
		o[2] = w->wa.y;
		o[4] = w->wa.h;
		break;
			
	case WF_PREVXYWH:				/* Get previous window position */
		o[1] = w->pr.x;
		o[2] = w->pr.y;
		o[3] = w->pr.w;
		o[4] = w->pr.h;
		break;			
			
	case WF_FULLXYWH:				/* Get maximum window dimensions */
		o[1] = root_window->r.x; 
		o[2] = root_window->wa.y;	/* Ensure the windows don't overlay the menu bar */
		o[3] = root_window->r.w;
		o[4] = root_window->wa.h;
		break;

	case WF_BEVENT:					/* HR */
		o[1] = ((w->active_widgets&NO_TOPPED) != 0)&1;
		break;	
	case WF_BOTTOM:					/* Extension, gets the bottom window */
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
#if 0
		for (w = window_list; w->next; w = w->next)
			;
#else
	/* HR!!! */
		w = root_window;
		if (w->prev)
			w = w->prev;
#endif
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);
		o[1] = w->handle;		/* Return the window handle of the bottom window */
		o[2] = w->owner;		/* Return the owner of the bottom window */
		break;
			
	case WF_TOP:
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);
		w = window_list;
		if (w)
		{
			o[1] = w->handle;	/* Return the window handle */
			o[2] = w->owner;	/* AES4 specifies that you return the AESid of the owner here as well */
			if (w->next)	/* Is there a window below?  */
			{				/* If there is, then AES4 says return its handle here */
				o[3] = w->next->handle;
				o[4] = w->next->owner;		/* XaAES extention - return the AESid of the app that owns the window below */
			} else
			{
				o[3] = 0;
				o[4] = 0;
			}
		} else
		{							/* Shouldnt we return the menu bar owner? */
			o[1] = 0;		/* No windows open - return an error */
			o[0] = 0;
		}
		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);		/* This should be at the end, because of prev, next */
		break;
			
	case WF_OWNER:					/* AES4 compatible stuff */
		if unlocked(winlist)
			Sema_Up(WIN_LIST_SEMA);					/* HR */

/* Little trace on root window owner
if (w == root_window && w->owner != 1)
	DIAG((D.z,-1,"HIER GAAT HET FOUT!!!!!\n"));
*/
		o[1] = w->owner;		/* The window owners AESid (==app_id) */
		o[2] = w->is_open;		/* Is the window open? */
		if (w->prev)	/* If there is a window above, return its handle */
			o[3] = w->prev->handle;
		else
			o[3] = -1;		/* HR: dont return a valid window handle (root window is 0) */
							/*     Some programs (Multistrip!!) get in a loop */
		if (w->next)	/* If there is a window below, return its handle */
			o[4] = w->next->handle;
		else
			o[4] = -1;

DIAG((D.z,-1,"own:%d, op:%d, pr:%d, nx:%d\n",o[1],o[2],o[3],o[4]));

		if unlocked(winlist)
			Sema_Dn(WIN_LIST_SEMA);					/* HR */
		break;

	case WF_VSLIDE:
		if (w->active_widgets&VSLIDE)
		{
			slw = w->widgets[XAW_VSLIDE].stuff;
			o[1] = slw->position;
		}
		else
			o[0] = o[1] = 0;
		break;
			
	case WF_HSLIDE:
		if (w->active_widgets&HSLIDE)
		{
			slw = w->widgets[XAW_HSLIDE].stuff;
			o[1] = slw->position;
		}
		else
			o[0] = o[1] = 0;
		break;
			
	case WF_HSLSIZE:
		if (w->active_widgets&HSLIDE)
		{
			slw = w->widgets[XAW_HSLIDE].stuff;
			o[1] = slw->length;
		}
		else
			o[0] = o[1] = 0;
		break;
			
	case WF_VSLSIZE:
		if (w->active_widgets & VSLIDE)
		{
			slw = w->widgets[XAW_VSLIDE].stuff;
			o[1] = slw->length;
		}
		else
			o[0] = o[1] = 0;
		break;

	case WF_SCREEN:		/* HR return a very small area :-) hope app's      */
	{					/*    then decide to allocate a buffer themselves. */
						/*    This worked for SELECTRIC.  */
						/*   Alas!!!! not PS_CONTRL.ACC	  */
						/* So now its all become official */
		extern char *half_screen_buffer;
		extern size_t half_screen_size;

		/* HR: make the half (or quarter) screen buffer for wind_get(WF_SCREEN) :-( :-( */
		/*    What punishment would be adequate for the bloke (or bird) who invented this? */

		if (half_screen_buffer == nil)
		{
			half_screen_size *= display.planes;
			half_screen_size /= display.planes < 8 ? 4 : 2;
			half_screen_buffer = Mxalloc(half_screen_size, MX_GLOBAL | MX_PREFTTRAM);
			client->half_screen = true;
		}

		*(char  **)&o[1] = half_screen_buffer;
		*(size_t *)&o[3] = half_screen_size;
		break;
	}
	case WF_NEWDESK:
		if unlocked(desk)
			Sema_Up(ROOT_SEMA);
		*(OBJECT **)&o[1] = desktop;
		if unlocked(desk)
			Sema_Dn(ROOT_SEMA);
		break;
	
	case WF_ICONIFY:
		if (w->window_status == XAWS_ICONIFIED)
		{
			o[1] = 1;
			o[2] = iconify_w;
			o[3] = iconify_h;
		}
		else
			o[1] = 0;
		break;

	case WF_UNICONIFY:
		o[1] = w->pr.x;
		o[2] = w->pr.y;
		o[3] = w->pr.w;
		o[4] = w->pr.h;
		break;

#ifdef __GNUC__
	case 0x5841:
		o[0] = 0x5841;
		break;
#else
	case 'XA':						/* HR: idee stolen from WINX */
		o[0] = 'XA';
		break;
#endif
	default:
		DIAG((D.wind,-1,"wind_get: %d\n",cmd));
		o[0] = 0;					/* HR */
	}
	
	return XAC_DONE;
}

short update_lock = 0;		/* HR: dont use FALSE/TRUE for numbers */
short mouse_lock = 0;
short update_cnt = 0;
short mouse_cnt = 0;

#define	EACCESS	36					/* Access denied */

/*
 * Wind_update handling
 *
 **  XA_wind_update runs under the client pid.
 *
 * This handles locking for the update and mctrl flags.
 * !!!!New version - uses semphores for locking...
 */

AES_function XA_wind_update	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	short op = pb->intin[0];
	long timeout = (op & 0x100) ? 0L : -1L;		/* Test for check-and-set mode */

	CONTROL(1,1,0)	

	pb->intout[0] = 1;

	switch(op)
	{
	case BEG_UPDATE:				/* Grab the update lock */
	case BEG_UPDATE|0x100:
		if (update_lock == client->pid)	/* Already owning it? */
			update_cnt++ ;
		else
		if (Psemaphore(2, UPDATE_LOCK, timeout) == -EACCESS)
			pb->intout[0] = 0;		/* Screen locked by different process */
		else
			update_lock = client->pid,
			update_cnt = 1 ;
		break;
	case END_UPDATE:
		if ((update_lock == client->pid) && (--update_cnt == 0))
		{
			update_lock = 0;
			Sema_Dn(UPDATE_LOCK);
		}
		break;
	case BEG_MCTRL:					/* Grab the mouse lock */
	case BEG_MCTRL|0x100:
		if (mouse_lock == client->pid)	/* Already owning it? */
		{
			mouse_cnt++ ;
			break ;
		}
		if (Psemaphore(2, MOUSE_LOCK, timeout) == -EACCESS)
		{
			pb->intout[0] = 0;		/* Mouse locked by different process */
			break ;
		}
		mouse_lock = client->pid;
		mouse_cnt = 1 ;
		break;
	case END_MCTRL:
		if ((mouse_lock == client->pid) && (--mouse_cnt == 0))
		{
			mouse_lock = FALSE;
			Sema_Dn(MOUSE_LOCK);
		}
		break;
	}
	return XAC_DONE;
}

AES_function XA_wind_delete	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	XA_WINDOW *w;
	
	CONTROL(1,1,0)	

	w = get_wind_by_handle(lock, pb->intin[0]);

	if (w and w->is_open == false)		/* HR: check if closed */
	{
		delete_window(lock, w);
		pb->intout[0] = 1;
	}
	else
		pb->intout[0] = 0;
	
	return XAC_DONE;
}

/*
 * Go through and check that all windows belonging to this client are
 * closed and deleted
 */

short wind_new(LOCK lock, short pid)
{
	XA_WINDOW *wl, *nwl;
	short owner_top_wm = -1;

	if unlocked(winlist)
		Sema_Up(WIN_LIST_SEMA);		/* HR */

	wl = window_list;
#if TWOLISTS
	while(wl != root_window)
#else
	while(wl)
#endif
	{
		nwl = wl->next;
		
		if (wl->owner == pid)
		{
#if ! TWOLISTS
			if (wl != root_window)
#endif
			{
				close_window(lock|winlist, wl);		/* checks is_open */
				delete_window(lock|winlist, wl);
			}
		} else
		{
			if ((owner_top_wm == -1) && Pid2Client(wl->owner)->std_menu.tree)	/* HR tree */
				owner_top_wm = wl->owner;
		}
		wl = nwl;
	}

#if TWOLISTS
	wl = closed_windows.first;
	while (wl)
	{
		nwl = wl->next;
		if (wl->owner == pid)
			delete_window(lock|winlist, wl);			
		wl = nwl;
	}
#endif
	
	if unlocked(winlist)
		Sema_Dn(WIN_LIST_SEMA);
	return 	owner_top_wm;
}

AES_function XA_wind_new	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(0,0,0)	

	wind_new(lock, client->pid);

	return XAC_DONE;
}


/*
 * wind_calc			/* HR: embedded function calc_window() */
 */
AES_function XA_wind_calc	/* (LOCK lock, XA_CLIENT *client, AESPB *pb) */
{
	CONTROL(6,5,0)	

	calc_window( client->pid,
				 pb->intin[0],		/* request */
				 pb->intin[1],		/* widget mask */
		(RECT *)&pb->intin[2],
		(RECT *)&pb->intout[1]);

	pb->intout[0] = 1;
	return XAC_DONE;
}
